/*   E-scape
     Copyright (C) 1997 Free Software Foundation, Inc.
     
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
     as published by the Free Software Foundation; either version 2
     of the License, or (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */


#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>

#include <config.h>

#include <stdio.h>
#include "libit/unistd.h"
#include "libtext.h"
#include "libnet.h"
#include "NetP.h"
#include "libit/malloc.h"

static void
_resolve_inheritance (WidgetClass class);

static void
initialize (Widget request, NetWidget self, ArgList args, Cardinal *num_args);

static int
set_values (Widget old, Widget request, NetWidget self, ArgList args,
	    Cardinal *num_args);

static void
destroy (NetWidget self);

static void
history_add (NetWidget self, char *url, char *post_data, size_t post_count);

static void
history_goto (NetWidget self, int no_cache);

static void
font_info (NetWidget self, struct css_properties *properties, float *em_factor,
	  float *ex_factor, void **font_opaque);

static void
app_data_html(struct E_scape_filter *filter, char *buf, size_t count);

static void
app_data(struct  E_scape_filter *filter, char *buf, size_t count);

static void
status_func (struct net_connection *net, char *mesg);

/* Add `url' with `post_data' of length `post_count' to the history
   stored in `self'. */
static void
history_add(NetWidget self, char *url, char *post_data, size_t post_count)
{
  struct history_struct *tmp, *tmp2;

#if 0
  if (!strcmp(url, self->Net.history->url))
    return;
#endif

  if (self->Net.history) {
    tmp = self->Net.history->next;
    while (tmp) {
      tmp2 = tmp->next;
      free((void *)tmp);
      tmp = tmp2;
    }
    self->Net.history->next = malloc(sizeof(struct history_struct));
    self->Net.history->next->back = self->Net.history;
    self->Net.history = self->Net.history->next;
  } else {
    self->Net.history = malloc(sizeof(struct history_struct));
    self->Net.history->back = 0;
  }
  self->Net.history->next = 0;
  self->Net.history->url = strdup(url);
  self->Net.history->title = strdup(url);
  self->Net.history->post_count = post_count;
  if (post_data) {
    self->Net.history->post_data = malloc(post_count);
    memcpy(self->Net.history->post_data, post_data, post_count);
  } else {
    self->Net.history->post_data = 0;
  }
}

/* Load the page pointed to by `self's history in `self'; don't use the
   cache if `no_cache' is nonzero.  */
/* FIXME this should store some state so that images in the page won't
   be loaded from the cache. */
static void
history_goto(NetWidget self, int no_cache)
{
  int read_condition, write_condition;
  XtAppContext context;

  read_condition = XtInputReadMask;
  write_condition = XtInputWriteMask;
  context = XtWidgetToApplicationContext(self);

  XtCallCallbackList(self, self->Net.activate, self->Net.history->url);

  if (self->Net.net)
    E_scape_net_free(self->Net.net);
  self->Net.net = E_scape_net_init(app_data, context, read_condition,
				   write_condition);
  self->Net.net->opaque = self;
  self->Net.net->status_func = status_func;
  if (self->Net.state) {
#if 0
    /* FIXME write this function and reenable this code */
    free_html_parser_state(self->Net.state);
#endif
  }

  self->Net.state = new_html_parser_state();
  self->Net.state->self = self;

  E_scape_net_goto(self->Net.net, self->Net.history->url,
		   self->Net.history->post_data, self->Net.history->post_count,
		   no_cache);
}



static void
display_tag (struct html_parser_state *x)
{
  /* FIXME */
}

static void
font_info (NetWidget self, struct css_properties *properties, float *em_factor,
	  float *ex_factor, void **font_opaque)
{
  /* FIXME these are the wrong parameters */
  *font_opaque = css1_font_to_x_font (XtDisplay(self), properties);
  /* FIXME this should check all of the fonts, in case the first
     one doesn't have the necissary character. */
  /* FIXME assign the correct values for em_factor and ex_factor */
}

/* Called whenever a new tag appears.  Fills in CSS1 info; if
   display-level, then check whether previous tag should be
   encapsulated in <P> and displayed.  If display-level, then
   display the contents of tag.  Also gets called recusively
   when the size of an image is discovered or a new style sheet
   becomes available.  */
void tag_done(struct html_parser_state *x)
{
  float medium_base = 10;
  float medium_factor = 1.2;
  NetWidget self = x->self;
  Screen *screen;

  if (x->current->parent) {
#if 0
    x->current->properties->parent = x->current->parent->properties;
    /* FIXME we don't need it unless it gets read somewhere... */
#endif
    /* FIXME medium_base and medium_factor should be intelligently
       initialized */
    screen = XtScreen(self);
    /* FIXME hardcoding "text/css" here is bogus */
    /* FIXME provide some real style sheets! */
    /* FIXME provide real pseudoclass handling */
    css_cascade_and_inherit (x->current, "text/css", 0, self, font_info,
			     screen->width / screen->mwidth,
			     screen->height / screen->mheight,
			     medium_base, medium_factor, 0, 0, 0, 0);
  } else {
    /* FIXME this is equally bogus */
    x->current->properties = default_css_properties();
  }
  switch (x->current->properties->display) {
  case DISPLAY_BLOCK:
    /* FIXME maybe need to put stuff into <P> */
#if 0
    display_tag(x); /* FIXME need to find out dims etc */
#endif
    break;
  case DISPLAY_LIST_ITEM: break; /* FIXME */
  case DISPLAY_INLINE: break;
  case DISPLAY_NONE: break;
  }
}

static void
app_data_html(struct E_scape_filter *filter, char *buf, size_t count)
{
  NetWidget self = filter->net->opaque;

  parse_html(self->Net.state, buf, count);
}

static void app_data(struct E_scape_filter *filter, char *buf, size_t count)
{
  NetWidget self = filter->net->opaque;

  if (!strcasecmp(filter->s, "image")) {
    /* FIXME */
  } else {
    if (!strcasecmp(filter->s, "text/plain")) {
      parse_html(self->Net.state, "<PLAINTEXT>", 11);
    }
    parse_html(self->Net.state, buf, count);
    filter->app_data = app_data_html;
  }
}

static void
status_func (struct net_connection *net, char *mesg)
{
  NetWidget self;
  self = net->opaque;
  XtCallCallbackList (self, self->Net.info, mesg);
}

static XtResource resources[] = {
{XtNactivate,XtCActivate,XtRCallback,sizeof(((NetRec*)NULL)->Net.activate),XtOffsetOf(NetRec,Net.activate),XtRImmediate,(XtPointer)NULL },
{XtNdone,XtCDone,XtRCallback,sizeof(((NetRec*)NULL)->Net.done),XtOffsetOf(NetRec,Net.done),XtRImmediate,(XtPointer)NULL },
{XtNtitle,XtCTitle,XtRCallback,sizeof(((NetRec*)NULL)->Net.title),XtOffsetOf(NetRec,Net.title),XtRImmediate,(XtPointer)NULL },
{XtNinfo,XtCInfo,XtRCallback,sizeof(((NetRec*)NULL)->Net.info),XtOffsetOf(NetRec,Net.info),XtRImmediate,(XtPointer)NULL },
};

NetClassRec netClassRec = {
{ /* core_class part */
/* superclass   	*/  (WidgetClass) &compositeClassRec,
/* class_name   	*/  "Net",
/* widget_size  	*/  sizeof(NetRec),
/* class_initialize 	*/  NULL,
/* class_part_initialize*/  _resolve_inheritance,
/* class_inited 	*/  FALSE,
/* initialize   	*/  initialize,
/* initialize_hook 	*/  NULL,
/* realize      	*/  XtInheritRealize,
/* actions      	*/  NULL,
/* num_actions  	*/  0,
/* resources    	*/  resources,
/* num_resources 	*/  4,
/* xrm_class    	*/  NULLQUARK,
/* compres_motion 	*/  False ,
/* compress_exposure 	*/  FALSE ,
/* compress_enterleave 	*/  False ,
/* visible_interest 	*/  False ,
/* destroy      	*/  destroy,
/* resize       	*/  XtInheritResize,
/* expose       	*/  XtInheritExpose,
/* set_values   	*/  set_values,
/* set_values_hook 	*/  NULL,
/* set_values_almost 	*/  XtInheritSetValuesAlmost,
/* get_values+hook 	*/  NULL,
/* accept_focus 	*/  XtInheritAcceptFocus,
/* version      	*/  XtVersion,
/* callback_private 	*/  NULL,
/* tm_table      	*/  NULL,
/* query_geometry 	*/  XtInheritQueryGeometry,
/* display_acceleator 	*/  XtInheritDisplayAccelerator,
/* extension    	*/  NULL 
},
{ /* composite_class part */
XtInheritGeometryManager,
XtInheritChangeManaged,
XtInheritInsertChild,
XtInheritDeleteChild,
NULL
},
{ /* Net_class part */
 /* dummy */  0
},
};

WidgetClass netWidgetClass = (WidgetClass) &netClassRec;

static void
_resolve_inheritance(WidgetClass class)
{
  NetWidgetClass c = (NetWidgetClass) class;
  NetWidgetClass super;
  static CompositeClassExtensionRec extension_rec = {
    NULL, NULLQUARK, XtCompositeExtensionVersion,
    sizeof(CompositeClassExtensionRec), True};
  CompositeClassExtensionRec *ext;
  ext = malloc(sizeof(*ext));
  *ext = extension_rec;
  ext->next_extension = c->composite_class.extension;
  c->composite_class.extension = ext;
  if (class == netWidgetClass) return;
  super = (NetWidgetClass)class->core_class.superclass;
}

static void
initialize(Widget request, NetWidget self, ArgList args, Cardinal *num_args)
{
  XGCValues values;

  self->Net.net = 0;
  /* FIXME check that we pass the right parameters */
  self->Net.gc = XtAllocateGC(self, self->core.depth, 0, &values,
			      GCForeground | GCBackground, 0);
}

static int
set_values(Widget old, Widget request, NetWidget self, ArgList args,
	   Cardinal *num_args)
{
  return 0;
}

static void
destroy(NetWidget self)
{
  if (self->Net.net)
    E_scape_net_free(self->Net.net);
}

void
NetForward(NetWidget self, NetWidget self2, XtPointer call_data)
{
  if (self2)
    self = self2;

  if ((self->Net.history) && (self->Net.history->next)) {
    self->Net.history = self->Net.history->next;
    history_goto(self, 0);
  }
}

void
NetBack(NetWidget self, NetWidget self2, XtPointer call_data)
{
  if (self2)
    self = self2;

  if ((self->Net.history) && (self->Net.history->back)) {
    self->Net.history = self->Net.history->back;
    history_goto(self, 0);
  }
}

void
NetHome(NetWidget self, NetWidget self2, XtPointer call_data)
{
  if (self2)
    self = self2;

  if (self->Net.history) {
    while (self->Net.history->back)
      self->Net.history = self->Net.history->back;
    history_goto(self, 0);
  }
}

void
NetReload(NetWidget self, NetWidget self2, XtPointer call_data)
{
  if (self2)
    self = self2;

  if (self->Net.history)
    history_goto(self, 1);
}


void
NetStop(NetWidget self, NetWidget self2, XtPointer trash)
{
  if (self2)
    self = self2;

  XtCallCallbackList(self, self->Net.done, self);
  if (self->Net.net)
    E_scape_net_stop(self->Net.net);
}

void
NetGoto(NetWidget self, char *URL, void *post_data, size_t post_count)
{
  history_add(self, URL, post_data, post_count);
  history_goto(self, 0);
}
