/*   E-scape authentication
     Copyright (C) 1997 Free Software Foundation, Inc.
     
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
     as published by the Free Software Foundation; either version 2
     of the License, or (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

/* FTP and HTTP both provide mechanisms for user authentication.  Generally,
 * they require a username and password.  For FTP, the hostname is the
 * only significant piece of information, but HTTP also uses a realm to
 * qualify the hostname.
 *
 * When E-scape needs authentication information, it will return a form
 * to the user, which will allow a username and password to be entered.
 * When the user submits the form, it will be sent to a special
 * authenticate://  After storing the authentication
 * information, a redirect will occur to display the original file.
 *
 * Misfeatures:
 *	The user can edit the hostname, protocol, etc, and mess things up.
 *		I need to look at the persistent cookie specs and clean
 *		this up
 *	Lots of password data can potentially pile up.  Also when you enter
 *		a new password for a given host/realm/protocol triplet, the
 *		old one is not forgotten.
 */

#include <config.h>
#include <noxt.h>
#include <libnet.h>

#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#endif

#include <libintl.h>
#define _(String) gettext(String)

#define NETWRITE(net, bar) \
	(net)->filter->app_data((net)->filter, (bar), strlen(bar))

void _E_scape_net_make_auth_form(struct net_connection *net)
{
	NETWRITE(net, _("<FORM ACTION=\"auth://\" METHOD=\"POST\">"
		"<TITLE>Authentication</TITLE>"
		"<H1>Authentication</H1>"
		"Protocol: <INPUT TYPE=\"text\" NAME=\"P\" VALUE=\""));
	NETWRITE(net, net->auth_protocol);
	NETWRITE(net,
		_("\"><P>Host: <INPUT TYPE=\"text\" NAME=\"H\" VALUE=\""));
	NETWRITE(net, net->url.hostname);
	NETWRITE(net,
		_("\"><P>Realm: <INPUT TYPE=\"text\" NAME=\"R\" VALUE=\""));
	NETWRITE(net, net->auth_realm);
	NETWRITE(net,
		_("\"><P>Name: <INPUT TYPE=\"text\" NAME=\"n\" VALUE=\""));
	if (net->url.username)
		NETWRITE(net, net->url.username);
	NETWRITE(net, _("><P>Password: "
		"<INPUT TYPE=\"password\" NAME=\"p\"><P>"
		"<INPUT TYPE=\"submit\" VALUE=\"OK\"><P></FORM>"));
	E_scape_net_stop(net);
}

struct auths {
	char *protocol, *host, *realm, *name, *password;
	struct auths *next;
};

static struct auths *auth;

/* A note to the paranoid: we do overwrite your password.  However,
 * it could concievably be swapped out for a short time...and
 * remain on the disk.  Or it could be swapped out, and the memory it
 * occupied be reassigned without being overwritten.
 */
void E_scape_net_auth_forget(void)
{
	struct auths *tmp;
	char *s;

	while (auth) {
		s = auth->password;
		while (*s) {
			*s = 0;
			s++;
		}

		free(auth->protocol);
		free(auth->host);
		free(auth->realm);
		free(auth->name);
		free(auth->password);

		tmp = auth->next;
		free(auth);
		auth = tmp;
	}
}

void E_scape_net_parse_auth(struct net_connection *net)
{
	char *s, *r, *protocol = 0, *host = 0, *realm = 0, *name = 0,
		*password = 0, **target;
	int count;
	struct auths *tmp;

	s = net->post_data;
	count = net->post_size;
	if (!s) count = 0;

	while (count && (*s == '?'))
		s++, count--;

	while (count) {
		while (count && (*s == '&'))
			s++, count--;
		if ((count > 2) && (s[1] == '=')) {
			target = 0;
			switch(*s) {
				case 'n':
					target = &name;
					break;
				case 'p':
					target = &password;
					break;
				case 'P':
					target = &protocol;
					break;
				case 'H':
					target = &host;
					break;
				case 'R':
					target = &realm;
					break;
			}
			if (target && (!*target)) {
				s++, count--;
				s++, count--;
				r = s;
				while (count && (*s != '&'))
					s++, count--;
				*target = malloc(s - r + 1);
				strncpy(*target, r, s - r);
				(*target)[s - r] = 0;
			}
		}
		while (count && (*s != '&'))
			s++, count--;
	}
	if (name && password && host && realm && protocol) {
		tmp = auth;
		auth = malloc(sizeof(struct auths));
		auth->name = name;
		auth->password = password;
		auth->host = host;
		auth->realm = realm;
		auth->protocol = protocol;
		auth->next = tmp;
	} else {
		if (name)
			free(name);
		if (password)
			free(password);
		if (host)
			free(host);
		if (realm)
			free(realm);
		if (protocol)
			free(protocol);
	}
}


void E_scape_net_auth_lookup(char *protocol, char *host, char *realm,
	char **name, char **password)
{
	struct auths *q;

	q = auth;
	while (q && (strcmp(protocol, q->protocol) || strcmp(host, q->host)
		|| strcmp(realm, q->realm)))
			q = q->next;
	if (q) {
		*name = q->name;
		*password = q->password;
	}
}
