/*   E-scape
     Copyright (C) 1997 Free Software Foundation, Inc.
     
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
     as published by the Free Software Foundation; either version 2
     of the License, or (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include <config.h>
#include "noxt.h"
#include "libnet.h"
#include "libit/unistd.h"

#define FREEFD(foo) if ((foo) != -1) {close((foo)); (foo) = -1; }

#ifndef FD_SETSIZE
#define FD_SETSIZE 256
#endif

static void cleanup(struct E_scape_filter *filter)
{
	FREEFD(filter->stdin_fd);
	FREEFD(filter->stdout_fd);
	FREEFD(filter->stderr_fd);
	if (filter->stdout_id) {
		XtRemoveInput(filter->stdout_id);
		filter->stdout_id = 0;
	}
	if (filter->stderr_id) {
		XtRemoveInput(filter->stderr_id);
		filter->stderr_id = 0;
	}
}

static void stdout_func(struct E_scape_filter *filter, int *fd, XtInputId *id)
{
	char buf[65536];
	size_t count;

	count = read(*fd, buf, 65536);
	if (!count) {
		cleanup(filter);
		return;
	}
	filter->net->app_data(filter->next, buf, count);
}


static void stderr_func(struct E_scape_filter *filter, int *fd, XtInputId *id)
{
	char buf[4097];
	size_t count;

	count = read(*fd, buf, 4096);
	if (!count) {
		cleanup(filter);
		return;
	}
	if (filter->net->status_func) {
		buf[count] = 0;
		filter->net->status_func(filter->net, buf);
	}
}


void _E_scape_net_child_filter(struct E_scape_filter *filter, char *buf,
        size_t count)
{
	int pipe_stdin[2], pipe_stdout[2], pipe_stderr[2], i;

	if (!buf) {
		/* Don't clean up, because the child may not be done yet. */
		return;
	}

	if (filter->stdin_fd == -1) {
		pipe(pipe_stdin);
		pipe(pipe_stdout);
		pipe(pipe_stderr);

		if (!fork()) {
			close(0);
			dup(pipe_stdin[0]);
			close(1);
			dup(pipe_stdout[1]);
			close(2);
			dup(pipe_stderr[2]);
			/* FIXME this isn't the right way to close all fd's */
			for (i = 3; i < FD_SETSIZE; i++)
				close(i);
			execl("/bin/sh", "sh", "-c", filter->s, 0);
		}
		close(pipe_stdin[0]);
		close(pipe_stdout[1]);
		close(pipe_stderr[1]);
		filter->stdin_fd = pipe_stdin[1];
		filter->stdout_fd = pipe_stdout[0];
		filter->stderr_fd = pipe_stderr[0];
		filter->stdout_id = XtAppAddInput(filter->net->app_context,
			filter->stdout_fd, filter->net->read_condition,
			(XtInputCallbackProc) stdout_func, filter);
		filter->stderr_id = XtAppAddInput(filter->net->app_context,
			filter->stderr_fd, filter->net->read_condition,
			(XtInputCallbackProc) stderr_func, filter);
	}

	full_write(filter->stdin_fd, buf, count);
}
