/*   E-scape css utils
     Copyright (C) 1997 Free Software Foundation, Inc.
     
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
     as published by the Free Software Foundation; either version 2
     of the License, or (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include <config.h>
#include <sys/types.h>
#include "libtext.h"
#include "libit/string.h"
#include "libit/malloc.h"

void copy_css_color(struct css_color *dest, struct css_color *src)
{
	dest->red = src->red;
	dest->green = src->green;
	dest->blue = src->blue;
	if (src->name)
		dest->name = strdup(src->name);
	else
		dest->name = 0;
}

void clear_css_color(struct css_color *tmp)
{
	tmp->red = tmp->green = tmp->blue = -1;
	tmp->name = 0;
}

struct css_properties *new_css_properties(void)
{
	struct css_properties *tmp;

	tmp = malloc(sizeof(struct css_properties));

	tmp->font_family = 0;
	tmp->font_style = 0;
	tmp->font_variant = 0;
	tmp->font_weight = 0;
	tmp->font_size.units = 0;
	clear_css_color(&tmp->color);
	clear_css_color(&tmp->background_color);
	tmp->background_url = 0;
	tmp->background_repeat = 0;
	tmp->background_scroll = 0;
	tmp->background_position_x.units = 0;
	tmp->background_position_y.units = 0;
	tmp->word_spacing.units = 0;
	tmp->letter_spacing.units = 0;
#if 0
	clear_css_color(&tmp->underline_color);
#endif
	tmp->text_decoration = 0;
	tmp->vertical_align.units = 0;
	tmp->text_indent.units = 0;
	tmp->text_transform = 0;
	tmp->text_align = 0;
	tmp->line_height.units = 0;
	tmp->margin_top.units = 0;
	tmp->margin_bottom.units = 0;
	tmp->margin_left.units = 0;
	tmp->margin_right.units = 0;
	tmp->padding_top.units = 0;
	tmp->padding_bottom.units = 0;
	tmp->padding_left.units = 0;
	tmp->padding_right.units = 0;
	clear_css_color(&tmp->border_top.color);
	clear_css_color(&tmp->border_bottom.color);
	clear_css_color(&tmp->border_left.color);
	clear_css_color(&tmp->border_right.color);
	tmp->border_top.width.units = 0;
	tmp->border_bottom.width.units = 0;
	tmp->border_left.width.units = 0;
	tmp->border_right.width.units = 0;
	tmp->border_top.style = 0;
	tmp->border_bottom.style = 0;
	tmp->border_left.style = 0;
	tmp->border_right.style = 0;
	tmp->width.units = 0;
	tmp->height.units = 0;
	tmp->float_mode = 0;
	tmp->clear = 0;
	tmp->display = 0;
	tmp->list_style_image_url = 0;
	tmp->list_style_type = 0;
	tmp->list_style_position = 0;
	tmp->white_space = 0;

	return tmp;
}

struct css_properties *default_css_properties(void)
{
	struct css_properties *tmp;

	tmp = malloc(sizeof(struct css_properties));

	tmp->font_family = 0; /* FIXME */
	tmp->font_style = FONT_STYLE_NORMAL;
	tmp->font_variant = FONT_VARIANT_NORMAL;
	tmp->font_weight = 500;
	tmp->font_size.units = 0; /* FIXME */
	clear_css_color(&tmp->color);
	clear_css_color(&tmp->background_color);
	tmp->background_url = 0;
	tmp->background_repeat = BACKGROUND_REPEAT_DEFINED;
	tmp->background_scroll = BACKGROUND_SCROLL;
	tmp->background_position_x.units = 0; /* FIXME */
	tmp->background_position_y.units = 0; /* FIXME */
	tmp->word_spacing.units = 0; /* FIXME */
	tmp->letter_spacing.units = 0; /* FIXME */
#if 0
	clear_css_color(&tmp->underline_color);
#endif
	tmp->text_decoration = TEXT_DECORATION_DEFINED;
	tmp->vertical_align.units = 0; /* FIXME */
	tmp->text_indent.units = 0; /* FIXME */
	tmp->text_transform = TRANSFORM_NONE;
	tmp->text_align = JUSTIFY_LEFT;
	tmp->line_height.units = 0; /* FIXME */
	tmp->margin_top.units = 0; /* FIXME */
	tmp->margin_bottom.units = 0; /* FIXME */
	tmp->margin_left.units = 0; /* FIXME */
	tmp->margin_right.units = 0; /* FIXME */
	tmp->padding_top.units = 0; /* FIXME */
	tmp->padding_bottom.units = 0; /* FIXME */
	tmp->padding_left.units = 0; /* FIXME */
	tmp->padding_right.units = 0; /* FIXME */
	clear_css_color(&tmp->border_top.color);
	clear_css_color(&tmp->border_bottom.color);
	clear_css_color(&tmp->border_left.color);
	clear_css_color(&tmp->border_right.color);
	tmp->border_top.width.units = 0; /* FIXME */
	tmp->border_bottom.width.units = 0; /* FIXME */
	tmp->border_left.width.units = 0; /* FIXME */
	tmp->border_right.width.units = 0; /* FIXME */
	tmp->border_top.style = 0; /* FIXME */
	tmp->border_bottom.style = 0; /* FIXME */
	tmp->border_left.style = 0; /* FIXME */
	tmp->border_right.style = 0; /* FIXME */
	tmp->width.units = 0; /* FIXME */
	tmp->height.units = 0; /* FIXME */
	tmp->float_mode = FLOAT_NONE;
	tmp->clear = 0; /* FIXME */
	tmp->display = DISPLAY_BLOCK;
	tmp->list_style_image_url = 0;
	tmp->list_style_type = 0; /* FIXME */
	tmp->list_style_position = 0; /* FIXME */
	tmp->white_space = WHITESPACE_NORMAL;

	return tmp;
}

/* Copy from `src' to `dest'.  If `use_factor' is non-zero, multiply
   the value by src->factor while copying it if appropraite.  */
void
copy_css_length(struct css_length *dest, struct css_length *src)
{
  dest->units = src->units;
  dest->value = src->value;
}

#define free_string(foo) if (foo) free(foo);


void free_css_selector(struct css_selector *sel)
{
	if (!sel) return;

	free_string(sel->tag);
	free_string(sel->class);
	free_string(sel->id);
	free_string(sel->pseudo);
	free_css_selector(sel->alt);
	free_css_selector(sel->parent);

	free(sel);
}

void free_css_properties(struct css_properties *properties)
{
	/*	FIXME write this correctly */
}

void free_css_rule(struct css_rule *rule)
{
	struct css_rule *tmp;

	while (rule) {
		free_css_selector(rule->sel);
		free_css_properties(rule->properties);
		tmp = rule->next;
		free(rule);
		rule = tmp;
	}
}

char **
copy_vector (char **src)
{
  char **tail, **dest, **x;

  tail = src;

  while (*tail)
    tail++;

  dest = malloc ((tail - src + 1) * sizeof (char **));

  x = dest;

  while (*src) {
    *x = strdup (*src);
    x++, src++;
  }

  *x = 0;

  return dest;
}


void
append_vector(char ***dest, char *src)
{
  size_t i;

  for (i = 0; (*dest)[i]; i++);

  *dest = realloc (dest, (i + 1) * sizeof (char **));

  (*dest)[i] = src;
  (*dest)[i + 1] = 0;
}


void free_vector(char **x)
{
  char **y;

  for (y = x; *y; y++)
    free (*y);
  free (x);
}
