/*   E-scape DNS child code
     Copyright (C) 1997 Free Software Foundation, Inc.
     
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
     as published by the Free Software Foundation; either version 2
     of the License, or (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include <config.h>

#include <netdb.h>
#include <netinet/in.h>
#include "libit/errno.h"
#include "libit/unistd.h"

#include "libit/malloc.h"

/* This file implements the DNS child process.  All it does is looks up
 * the hostname specified by the parent on stdin, and then either writes
 * errno to stderr on failure or the address of the host on stdout
 * on success.
 *
 * I'd prefer to have this code in the parent, as all the parent/child
 * synchronization effort is a pain.  However, including full DNS lookup
 * code in E-scape would be even more of a pain, and as the current library
 * call is coded,
 * DNS lookups block their process.  Maybe on Hurd we might be able
 * to abandon the multiple process technique...but even then backwards
 * compatibility might be important.
 *
 * Someday I may go for multiple name lookups for one process, but for
 * now we won't.
 */

/* We assume the parent was kind enough to null terminate the hostname.
 */

extern int h_errno;

int main(int argc, char *argv[])
{
	char *tmp;
	int i;
	int tmpsize, count, pos;
	struct hostent *host;

	tmpsize = 256;
	tmp = malloc(tmpsize);

	pos = 0;
	do {
		count = read(0, tmp + pos, tmpsize - pos);
		if (count < 0)
			return 1;
		pos += count;
		if (pos == tmpsize) {
			tmpsize *= 2;
			tmp = realloc(tmp, tmpsize);
		}
	} while (count > 0);

	host = gethostbyname(tmp);
	if (host) {
		write(1, host->h_addr, sizeof(struct in_addr));
	} else {
		i = h_errno;
		write(2, &i, sizeof(i));
	}
	return 0;
}
