/*   E-scape
     Copyright (C) 1997 Free Software Foundation, Inc.
     
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
     as published by the Free Software Foundation; either version 2
     of the License, or (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

/* The general goal of this program is to be a driver for testing the net
 * code.  It will also be a command line utility for downloading data
 * from an arbitrary URL.  It may someday be involved in allowing you
 * to do something like
 * less ftp://devnull@wombat.gnu.ai.mit.edu/com/mailer/aliases
 * on hurd.  Actually, I think I should bundle a lessopen file that
 * uses this program so that that sort of trick becomes possible.
 *
 * Note that in the above example, I may not have gotten the exact
 * path right.  Also, I have no idea how password handling will work
 * in this case.  Less will think it owns the tty.  And I believe
 * anonymous ftp @gnu is chrooted.  Maybe this example will change
 * their mind...they don't like security when it restricts useful
 * activites.
 *
 * Alternatively, you could include the password:
 * less ftp://devnull:censored@wombat.gnu.ai.mit.edu/path
 * (and censored is not my password, and never has been, and definately
 *  never will be...)
 * however if you do this people can learn your password with the `ps'
 * or 'w' command.  This is particularily dangerous on sites like FSF
 * that have finger daemons that reveal what program you are running.
 */

#include <config.h>
#include <version.h>
#include "noxt.h"
#include "libnet.h"
#include "stdio.h"
#include "libit/getopt.h"
#include "libit/unistd.h"

#include <libintl.h>
#define _(String) gettext(String)

char *program_name;

int bytes_transfered;

static int version = 0, help = 0, verbose = 1;

static struct option longopts[] = {
  {"version",	no_argument,	&version,	1},
  {"help",	no_argument, 	&help,		1},
  {"verbose",	no_argument,	&verbose,	2},
  {"silent",	no_argument,	&verbose,	0},
  {"quiet",	no_argument,	&verbose,	0},
};

/* If `buf' is non-zero, copy `count' bytes at `buf' to `stdout'; otherwise
   call `exit'. */
void data_function(struct E_scape_filter *filter, char *buf, size_t count)
{
  if (buf)
    full_write(1, buf, count);
  else
    exit(0);
}

void status_func(struct net_connection *net, char *mesg)
{
  fprintf(stderr, "%s", mesg);
}

int main(int argc, char *argv[])
{
  int c;
  struct net_connection *net;
  XtAppContext app;

  program_name = argv[0];

  E_scape_zombie_init();

  do {
    c = getopt_long(argc, argv, "hv", longopts, 0);
    switch(c) {
    case 'v':
      version = 1;
      break;
    case 'h':
      help = 1;
      break;
    }
  } while (c != -1);

  if (help) {
    fprintf(stderr, _("E-scape-get: send data at URL to stdout\n"
		      "E-scape-get [-v] [-h] [--version] [--help]"
		      " [--silent] [--quiet] [--verbose] <URL>\n"
		      "-v, --version           Print version and exit\n"
		      "-h, --help              Print help and exit\n"
		      "    --silent, --quiet   Be quiet\n"
		      "    --verbose           Be verbose\n"));
  } else if (version) {
    fprintf(stderr, _("%s %s\n"), PACKAGE, VERSION);
  } else if (optind == argc) {
    fprintf (stderr, _("No URL specified\n"));
  } else if ((optind + 1) != argc) {
    fprintf (stderr, _("Too many parameters specified\n"));
  } else {
    E_scape_ident_init();
    app = NoXtInit();
    net = E_scape_net_init(data_function, app, XtInputReadMask,
			   XtInputWriteMask);
    if (verbose == 2)
      net->status_func = status_func;
    E_scape_net_goto(net, argv[optind], 0, 0, 0);
    XtAppMainLoop(app);
  }
  exit(0);
}
