/*   E-scape--file transfer protocol (rfc959)
     Copyright (C) 1997 Free Software Foundation, Inc.
     
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
     as published by the Free Software Foundation; either version 2
     of the License, or (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

/* This ftp implementation may be in slight violation of the specs:
 * my understanding is that the ability to use the default port is
 * REQUIRED by the spec.  However, it also appears that due to certain
 * security features in the Linux kernel, IT IS NOT POSSIBLE for the
 * default port to be used.
 *
 * Since my observations leave me with the impression that the PORT
 * command is always used, I'll just assume that this is safe.  If
 * you know of any problems, feel free to let me know.
 *
 * Another unfortunate reality of the real world is that there are several
 * possible formats for directory listings.  On some wierd servers you
 * may have problems; feel free to report hostnames that cause problems.
 *
 * The current assumption is that we can reasonably use the output of
 * the LIST command.  There's also an NLST command that returns only
 * filenames, but its output is less useful when talking to a UN*X system.
 *
 * An interesting test will be how well this code works with the ftp
 * server on linux.mit.edu
 *
 * Experimentation finds that the ftpd I have on nautilus (which is
 * the standard one shiped with slackware 3, so there are lots
 * of copies of it out there...) returns 550 when I try to RETR either
 * a directory or something that doesn't exist.  AI to parse the message?
 *
 * An additional problem is that in the current command line client,
 * when the data connection closes, the process dies.  This means
 * I don't see all the control messages...I don't see any solutions
 * to this problem at 20:41 HST.  (I suppose if I give the MA times
 * it would look like I stay up all night like a normal hacker, since
 * staying up all night isn't practical with school...)
 */

#include <config.h>
#include <noxt.h>
#include <libnet.h>
#include "libit/string.h"
#include "libit/unistd.h"
#include "libit/errno.h"
#include <sys/socket.h>
#include <netinet/in.h>
#include <stdio.h>

#define NETWRITE(bar) _E_scape_net_write(net, (bar), strlen(bar))

/* FIXME should do error checking */

void nullresponse(struct net_connection *net)
{
}

void getinfo(struct net_connection *net)
{
	char *s;

	switch (net->ftp_response_code) {
		case 150: /* FIXME check if this actually works */
			s = net->ftp_buf + 4;
			while (((*s < '0') || (*s > '9')) && (*s)) s++;
			if ((*s >= '0') && (*s <= '9')) {
				net->total_size = 0;
				while ((*s >= '0') && (*s <= '9')) {
					net->total_size *= 10;
					net->total_size += *s - '0';
					s++;
				}
			}
			net->ftp_response = nullresponse;
			break;
		case 550:
			NETWRITE("LIST ");
			NETWRITE(net->url.filename);
			NETWRITE("\r\n");
			net->mime_type = strdup("text/x-ftp-dir");
			break;
	}
}

void sendget(struct net_connection *net)
{
	NETWRITE("RETR ");
	NETWRITE(net->url.filename);
	NETWRITE("\r\n");
	net->ftp_response = getinfo;
}

void sendport(struct net_connection *net)
{
	NETWRITE(net->port_cmd);
	net->ftp_response = sendget;
}

void sendmode(struct net_connection *net)
{
	NETWRITE("TYPE I\r\n");
	net->ftp_response = sendport;
}

void sendpass(struct net_connection *net)
{
	NETWRITE("PASS ");
	/* FIXME use a real mail address */
	NETWRITE("chaos@gnu.ai.mit.edu");
	NETWRITE("\r\n");
	net->ftp_response = sendmode;
}

void senduser(struct net_connection *net)
{
	NETWRITE("USER ");
	if (net->url.username && (net->url.username != ""))
		NETWRITE(net->url.username);
	else
		NETWRITE("anonymous");
	NETWRITE("\r\n");
	net->ftp_response = sendpass;
}

static void ftp_response(struct net_connection *net, int *fd, XtInputId *id)
{
	char *s;
	int count;

	if (net->ftp_buf_size == net->ftp_buf_count) {
		net->ftp_buf_size *= 2;
		net->ftp_buf = realloc(net->ftp_buf, net->ftp_buf_size);
	}

	count = read(net->net_fd, net->ftp_buf + net->ftp_buf_count,
		net->ftp_buf_size - net->ftp_buf_count);

	if (count == -1) _E_scape_net_html_error(net,
		"ftp: control read", errno);

	if (!count) return;

	net->ftp_buf_count += count;

	/* Once we have the data, we have to check if it is a full response.
	 * The conditional is a cheap hack; by definition if don't
	 * have a newline, we don't have a full line.  However, there is
	 * such a thing as a continuation with a dash, and so we have
	 * to carefully check that posibility.
	 */
	if ((net->ftp_buf[net->ftp_buf_count - 1] != '\r')
			&& (net->ftp_buf[net->ftp_buf_count - 1] != '\n'))
		return;

	/* The following paragraph has the effect of printing the status
	 * on the screen.  The null byte will also guarentee that we
	 * don't segfault if we have tons of digits, but so does the
	 * check for a newline...
	 */
	net->ftp_buf[net->ftp_buf_count] = 0;
	if (net->status_func)
		net->status_func(net, net->ftp_buf);

	s = net->ftp_buf + net->ftp_buf_count - 3;
	while ((*s != '\n') && (*s != '\r') && (s > net->ftp_buf)) s--;
	if ((*s < '0') || (*s > '9')) s++;
	net->ftp_response_code = 0;
	while ((*s <= '9') && (*s >= '0')) {
		net->ftp_response_code *= 10;
		net->ftp_response_code += *s - '0';
		s++;
	}
	if (*s == '-') return;

#if 0
	switch (net->ftp_response_code / 100) {
		case 1: ;
		case 2:	;
		case 3:	net->ftp_response(net);
			break;
		default: _E_scape_net_html_error(net, net->ftp_buf, 0);
	}
#endif
	net->ftp_response(net);

	net->ftp_buf_count = 0;
}

static void ftp_listen(struct net_connection *net, int *fd, XtInputId *id)
{
	struct sockaddr_in name;
	size_t len;

	len = sizeof(name);
	net->data_fd = accept(net->listen_fd, (struct sockaddr *) &name, &len);
	if (net->data_fd == -1)
		_E_scape_net_html_error(net, "accept", errno);
	_E_scape_net_input_install(net);
}

void _E_scape_connected_ftp(struct net_connection *net)
{
	struct sockaddr_in name;
	size_t len;

	net->ftp_buf_size = 4096;
	net->ftp_buf = malloc(net->ftp_buf_size);
	net->ftp_buf_count = 0;

	XtRemoveInput(net->net_id);
	net->ftp_response = senduser;
	net->net_id = XtAppAddInput(net->app_context, net->net_fd,
		net->read_condition, (XtInputCallbackProc) ftp_response, net);

	net->listen_fd = socket(AF_INET, SOCK_STREAM, 0);
	if (net->listen_fd == -1)
		_E_scape_net_html_error(net, "listen", errno);

	/* I'm not sure if the following paragraph is nessisary, but
	 * the equivalent of this is present in the W3C code, and
	 * I've already had enough problems getting this code working
	 * today, so I'll put it in in case it is needed.  If you're
	 * bored try commenting it out and see if we really need it.
	 */
	name.sin_family = AF_INET;
	name.sin_addr.s_addr = INADDR_ANY;

	len = sizeof(name);
	if (getsockname(net->net_fd, (struct sockaddr *) &name, &len) < 0)
		_E_scape_net_html_error(net, "getsockname", errno);

	name.sin_port = 0;

	if (bind(net->listen_fd, (struct sockaddr *) &name, len) < 0)
		_E_scape_net_html_error(net, "bind", errno);

	len = sizeof(name);
	if (getsockname(net->listen_fd, (struct sockaddr *) &name, &len) < 0)
		_E_scape_net_html_error(net, "getsockname", errno);

	sprintf(net->port_cmd, "PORT %d,%d,%d,%d,%d,%d\r\n",
		(int)*((unsigned char *)(&name.sin_addr)+0),
		(int)*((unsigned char *)(&name.sin_addr)+1),
		(int)*((unsigned char *)(&name.sin_addr)+2),
		(int)*((unsigned char *)(&name.sin_addr)+3),
		(int)*((unsigned char *)(&name.sin_port)+0),
		(int)*((unsigned char *)(&name.sin_port)+1));

	if (listen(net->listen_fd, 1) < 0)
		_E_scape_net_html_error(net, "listen", errno);

	net->listen_id = XtAppAddInput(net->app_context, net->listen_fd,
		net->read_condition, (XtInputCallbackProc) ftp_listen, net);
}
