/*   E-scape--http protocol
     Copyright (C) 1997 Free Software Foundation, Inc.
     
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
     as published by the Free Software Foundation; either version 2
     of the License, or (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include <config.h>
#include <version.h>

#include "noxt.h"
#include "libnet.h"

#include "libit/string.h"
#include "libit/ctype.h"
#include "libit/unistd.h"

#include <libintl.h>
#define _(String) gettext(String)

#include "libit/malloc.h"

/* I am currently having feelings about the creator of HTTP which are less
 * than totally respectful.  I do not understand why MIME headers
 * have to have both a carriage return and a linefeed.  It wastes
 * bandwidth, and parsing it is more difficult.  There's a good reason
 * UN*X uses \n.  Of course Mr. Gates was pretty idiotic as well.
 */

#define STATE_STATUS 0
#define STATE_NEWLINE 1
#define STATE_FIELD 2
#define STATE_SEPARATOR 3
#define STATE_INFO 4
#define STATE_END 5

inline static void process_status(struct net_connection *net)
{
	if (!strncmp(net->mime_field, "HTTP/1.", 7)) {
#if 0
	if ((!strncmp(net->mime_field, "HTTP/1.0 ", 9))
			&& isdigit(net->mime_field[9])
			&& isdigit(net->mime_field[10])
			&& isdigit(net->mime_field[11])) {
#if 0
		net->http_status = 100 * (net->mime_field[9] - '0')
			+ 10 * (net->mime_field[10] - '0')
			+ (net->mime_field[11] - '0');
#endif
#endif
	} else {
		net->url.protocol = realloc(net->url.protocol, 9);
		strcpy(net->url.protocol, "http_0_9");
		_E_scape_net_go(net);
	}
}

inline static void process_pair(struct net_connection *net)
{
	char *s;

	if (!strcasecmp(net->mime_field, "Content-type")) {
		if (net->mime_type)
			free(net->mime_type);
		net->mime_type = net->mime_info;
		net->mime_info = 0;
	} else if (!strcasecmp(net->mime_field, "Content-encoding")) {
		if (net->mime_encoding)
			free(net->mime_encoding);
		net->mime_encoding = net->mime_info;
		net->mime_info = 0;
	} else if (!strcasecmp(net->mime_field, "Content-length")) {
		net->total_size = 0;
		s = net->mime_info;
		while (*s) {
			if (isdigit(*s)) {
				net->total_size *= 10;
				net->total_size += *s - '0';
			}
			s++;
		}
		free(net->mime_info);
		net->mime_info = 0;
	} else if (!strcasecmp(net->mime_field, "WWW-Authenticate")) {
		if (net->auth_challenge)
			free(net->auth_challenge);
		net->auth_challenge = net->mime_info;
		net->mime_info = 0;
	} else if (!strcasecmp(net->mime_field, "Location")) {
		if (net->location)
			free(net->location);
		net->location = net->mime_info;
		net->mime_info = 0;
	} else {
		free(net->mime_info);
		net->mime_info = 0;
	}

	free(net->mime_field);
	net->mime_field = 0;
}


inline static void *find_nlcr(const void *block, size_t size)
{
	void *nl, *cr;

	nl = memchr(block, '\n', size);
	cr = memchr(block, '\r', size);

	if ((!nl) && (!cr))
		return 0;

	if ((nl) && (!cr))
		return nl;

	if ((!nl) && (cr))
		return cr;

	if (nl < cr)
		return nl;

	return cr;
}

static void append(char **dest, char *src, size_t size)
{
	size_t len;

	if (*dest) {
		len = strlen(*dest);
		*dest = realloc(*dest, size + len + 1);
		memcpy((*dest) + len, src, size);
		(*dest)[len + size] = 0;
	} else {
		*dest = malloc(size + 1);
		memcpy((*dest), src, size);
		(*dest)[size] = 0;
	}
}

/* There are a number of subtleties of the HTTP specs that rarely matter.
 * 204 No Content is not handled correctly.  Redirection can occur with
 * POST requests, but it shouldn't.  And I ignore them for now.
 *
 * In general, if we see a Location: header we redirect, if we get
 * WWW-authenticate, we send an authentication form, otherwise we
 * display what we get, since we usually get a nice HTML error message.
 */

inline static void header_done(struct net_connection *net)
{
	if (net->auth_challenge) {
		_E_scape_net_make_auth_form(net);
	} else if (net->location) {
		net->redir_count--;
		if (!net->redir_count)
			_E_scape_net_html_error(net,
				_("Too many redirections"), 0);
		_E_scape_net_url_parse(&net->url, net->location);
		free(net->location);
		net->location = 0;
		_E_scape_net_go(net);
	} else {
		XtRemoveInput(net->net_id);
		net->net_id = 0;
		_E_scape_net_input_install(net);
	}
}

static void parse_mime(struct net_connection *net, int *fd, XtInputId *id)
{
	char buf[65536];
	int count;
	char *s, *nl, *end;

	count = read(net->net_fd, &buf, 65536);
	s = (char *) &buf;
	end = s + count;
	while (s < end) {
		switch(net->mime_state) {
			case STATE_STATUS:
				nl = find_nlcr(s, end - s);
				if (nl) {
					append(&(net->mime_field), s, nl - s);
					process_status(net);
					free(net->mime_field);
					net->mime_field = 0;
					net->mime_state = STATE_NEWLINE;
					net->mime_nl = *nl;
					s = nl + 1;
				} else {
					append(&(net->mime_field), s, end - s);
					s = end;
				}
				break;
			case STATE_NEWLINE:
				if ((net->mime_nl == '\r') && (*s == '\n')) {
					s++;
					net->mime_nl = ' ';
				} else if ((*s == '\n') || (*s == '\r')) {
					net->mime_state = STATE_END;
					net->mime_nl = *s;
					s++;
				} else {
					net->mime_state = STATE_FIELD;
				}
				break;
			case STATE_FIELD:
				nl = memchr(s, ':', end - s);
				if (nl) {
					append(&(net->mime_field), s, nl - s);
					s = nl + 1;
					net->mime_state = STATE_SEPARATOR;
				} else {
					append(&(net->mime_field), s, end - s);
					s = end;
				}
				break;
			case STATE_SEPARATOR:
				while ((s != end) && (*s == ' '))
					s++;
				if (s != end)
					net->mime_state = STATE_INFO;
				break;
			case STATE_INFO:
				nl = find_nlcr(s, end - s);
				if (nl) {
					append(&(net->mime_info), s, nl - s);
					process_pair(net);
					net->mime_state = STATE_NEWLINE;
					net->mime_nl = *nl;
					s = nl + 1;
				} else {
					append(&(net->mime_field), s, end - s);
					s = end;
				}
				break;
			case STATE_END:
				if ((net->mime_nl == '\r') && (*s == '\n')) {
					s++;
					net->mime_nl = ' ';
				} else {
					net->filter->app_data(net->filter, s,
						end - s);
					s = end;
					header_done(net);
				}
				break;
		}
	}
}

void _E_scape_connected_http(struct net_connection *net)
{
	XtRemoveInput(net->net_id);
	net->net_id = 0;

	_E_scape_net_write(net, "GET ", 4);
	_E_scape_net_write(net, net->url.filename, strlen(net->url.filename));
	_E_scape_net_write(net, " HTTP/1.0\r\n", 11);

	/* FIXME authorization */

	/* I don't really want to remove the email address transmission
	 * code; however, I'm not sure that it is nessisary, either.
	 * The solution is to leave the #if 0 around it.  I wouldn't
	 * mind hearing from anyone who feels that email addresses
	 * need to be advertised.  But I don't see the benefit.  And
	 * proxies complicate the situation.
	 */
#if 0
	if (E_scape_email_http) {
		E_scape_net_write(net, "from: ", 6);
		E_scape_net_write(net, E_scape_email_http,
			strlen(E_scape_email_http));
		E_scape_net_write(net, "\r\n", 2);
	}
#endif

	if (net->no_cache) {
		_E_scape_net_write(net, "Pragma: no-cache\r\n", 18);
	}

	_E_scape_net_write(net, "user-agent: E-scape/", 20);
	_E_scape_net_write(net, VERSION, strlen(VERSION));
	_E_scape_net_write(net, "\r\n", 2);

	_E_scape_net_write(net, "\r\n", 2);

	net->net_id = XtAppAddInput(net->app_context, net->net_fd,
		net->read_condition, (XtInputCallbackProc) parse_mime, net);
}
