/*   E-scape file identification routines
     Copyright (C) 1997 Free Software Foundation, Inc.
     
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
     as published by the Free Software Foundation; either version 2
     of the License, or (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

/* In the begining, there was plain text.  And the hackers saw that it
 * was good.  Then they created graphics, and they saw that the pictures
 * were beautiful.  But as time continued, they realized that they had
 * different, incompatible ways of storing the same information.  In their
 * quest to solve the world's problems, MIME was invented as the savior of
 * this chaos.  Unfortunately, ftp had already been created without MIME
 * types.
 */

#include <config.h>
#include "noxt.h"
#include "libnet.h"
#include "libit/malloc.h"

#define MAX(foo, bar) ((foo) > (bar) ? (foo) : (bar))
#define COMPARE_STRING 0
#define COMPARE_START 1
#define COMPARE_ALL 2

static struct type_info *info;

/* The following four functions created with the clone-and-hack technique... */

inline static void add_mime(int priority, char *data)
{
	struct identify_info *foo;
	int bar;

	foo = malloc(sizeof(struct identify_info));
	foo->next = info->mime;
	info->mime = foo;

	foo->priority = priority;
	foo->length = strlen(data);
	foo->data = malloc(foo->length);
	foo->mask = malloc(foo->length);
	for (bar = 0; bar < foo->length; bar++) {
		foo->mask[bar] = 0xff;
		foo->data[bar] = data[bar];
	}
}

inline static void add_ext(int priority, char *data)
{
	struct identify_info *foo;
	int bar;

	foo = malloc(sizeof(struct identify_info));
	foo->next = info->ext;
	info->ext = foo;

	foo->priority = priority;
	foo->length = strlen(data);
	foo->data = malloc(foo->length);
	foo->mask = malloc(foo->length);
	for (bar = 0; bar < foo->length; bar++) {
		foo->mask[bar] = 0xff;
		foo->data[bar] = data[bar];
	}
}

inline static void add_common(int priority, char *data)
{
	struct identify_info *foo;
	int bar;

	foo = malloc(sizeof(struct identify_info));
	foo->next = info->string;
	info->string = foo;

	foo->priority = priority;
	foo->length = strlen(data);
	foo->data = malloc(foo->length);
	foo->mask = malloc(foo->length);
	for (bar = 0; bar < foo->length; bar++) {
		foo->mask[bar] = 0xff;
		foo->data[bar] = data[bar];
	}
}

inline static void add_magic(int priority, char *data)
{
	struct identify_info *foo;
	int bar;

	foo = malloc(sizeof(struct identify_info));
	foo->next = info->magic;
	info->magic = foo;

	foo->priority = priority;
	foo->length = strlen(data);
	foo->data = malloc(foo->length);
	foo->mask = malloc(foo->length);
	for (bar = 0; bar < foo->length; bar++) {
		foo->mask[bar] = 0xff;
		foo->data[bar] = data[bar];
	}
}

inline static void new_type(void *app_data, char *s)
{
	struct type_info *new;

	new = malloc(sizeof(struct type_info));
	new->next = info;
	info = new;
	info->mime = info->ext = info->magic = info->string = 0;
	info->app_data = app_data;
	info->s = s;
}

#define LONG_MAGIC_PRIORITY	4096
#define SHORT_MAGIC_PRIORITY	1024
#define MIME_PRIORITY		256
#define EXT_PRIORITY		256
#define HTML_TAG_PRIORITY	64

void E_scape_ident_init(void)
{
	/* Since image files are guarenteed to have one of these magic
	 * numbers if we know how to deal with them, we won't worry about
	 * other possible methods of recongizing them.
	 */
	new_type(0, "image");
	add_magic(LONG_MAGIC_PRIORITY, "GIF89a");
	add_magic(LONG_MAGIC_PRIORITY, "GIF87a");
	add_magic(SHORT_MAGIC_PRIORITY, "P1\n");
	add_magic(SHORT_MAGIC_PRIORITY, "P2\n");
	add_magic(SHORT_MAGIC_PRIORITY, "P3\n");
	add_magic(SHORT_MAGIC_PRIORITY, "P4\n");
	add_magic(SHORT_MAGIC_PRIORITY, "P5\n");
	add_magic(SHORT_MAGIC_PRIORITY, "P6\n");
	add_magic(SHORT_MAGIC_PRIORITY, "P1\r");
	add_magic(SHORT_MAGIC_PRIORITY, "P2\r");
	add_magic(SHORT_MAGIC_PRIORITY, "P3\r");
	add_magic(SHORT_MAGIC_PRIORITY, "P4\r");
	add_magic(SHORT_MAGIC_PRIORITY, "P5\r");
	add_magic(SHORT_MAGIC_PRIORITY, "P6\r");
	add_magic(SHORT_MAGIC_PRIORITY, "P1 ");
	add_magic(SHORT_MAGIC_PRIORITY, "P2 ");
	add_magic(SHORT_MAGIC_PRIORITY, "P3 ");
	add_magic(SHORT_MAGIC_PRIORITY, "P4 ");
	add_magic(SHORT_MAGIC_PRIORITY, "P5 ");
	add_magic(SHORT_MAGIC_PRIORITY, "P6 ");
	add_magic(LONG_MAGIC_PRIORITY, "\211PNG\r\n\032\n");

	/* We try to identify HTML by looking for common tags.  Since this
	 * is not a rock-solid method of identifcation anyway, and since
	 * people usually write all caps or or all lower case, we  don't
	 * worry about all possibilities.
	 */
	new_type(0, "text/html");
	add_common(HTML_TAG_PRIORITY, "<HTML>");
	add_common(HTML_TAG_PRIORITY, "<html>");
	add_common(HTML_TAG_PRIORITY, "<TITLE>");
	add_common(HTML_TAG_PRIORITY, "<title>");
	add_common(HTML_TAG_PRIORITY, "<HEAD>");
	add_common(HTML_TAG_PRIORITY, "<head>");
	add_common(HTML_TAG_PRIORITY, "<BODY>");
	add_common(HTML_TAG_PRIORITY, "<body>");
	add_common(HTML_TAG_PRIORITY, "<H1>");
	add_common(HTML_TAG_PRIORITY, "<h1>");
	add_mime(MIME_PRIORITY, "text/html");
	add_ext(EXT_PRIORITY, "html");
	add_ext(EXT_PRIORITY, "htm");

	/* Plain text is a fallback default.  Also, if someone tells us
	 * it's plain text, then we generally want to trust them.
	 */
	new_type(0, "text/plain");
	add_mime(MIME_PRIORITY, "text/plain");
	add_ext(EXT_PRIORITY, "txt");
	add_ext(EXT_PRIORITY, "text");
	add_ext(EXT_PRIORITY, "c");
	add_ext(EXT_PRIORITY, "el"); /* RMS has been recognized */
	add_ext(EXT_PRIORITY, "pas"); /* entity@koa has been remembered... */
	/* What was oyama@cco.caltech.edu's new extension? */
	/* Sorry Fran, but you speak C as well as I do... */
	/* What's the normal extension used by make menuconfig?
	 * roadcapw@cfw.com has been remembered...
	 */
	/* Okay, so using this section to say hi to my friends is wierd.
	 * but I like it...
	 */

	/* The following is a nice little hack to make text the default.
	 * it relies on some not-previously documented properties of
	 * magic number handling.  Since this comment is documentation,
	 * it is now documented.  ;-)
	 */
	add_magic(1, "");

	/* FIXME support an external file describing filtering */
#if 0
	new_type(_E_scape_net_child_filter, "gzip -cdq");
#endif
	new_type(_E_scape_net_gunzip_filter, 0);
	add_magic(SHORT_MAGIC_PRIORITY, "\037\235");
	add_magic(SHORT_MAGIC_PRIORITY, "\037\036");
	add_magic(SHORT_MAGIC_PRIORITY, "\037\213");
	add_magic(SHORT_MAGIC_PRIORITY, "\037\236");
	add_magic(SHORT_MAGIC_PRIORITY, "\037\240");

	new_type(_E_scape_net_deflate_filter, 0);
	add_mime(MIME_PRIORITY, "deflate");

	new_type(_E_scape_net_child_filter, "jpeg -pnm");
	add_magic(LONG_MAGIC_PRIORITY, "\377\330\377\340");

	new_type(_E_scape_net_child_filter, "bmptoppm 2>/dev/null");
	add_magic(SHORT_MAGIC_PRIORITY, "BM");

	new_type(_E_scape_net_child_filter, "rasttoppm 2>/dev/null");
	add_magic(LONG_MAGIC_PRIORITY, "\x59\xa6\x6a\x95");

	new_type(_E_scape_net_child_filter, "sgitopnm 2>/dev/null");
	add_magic(SHORT_MAGIC_PRIORITY, "\x1\xda");
	add_magic(SHORT_MAGIC_PRIORITY, "\xda\x1");
}

/* Like memcmp, but has extra param specifying a mask.  Probably not
 * stunningly efficent on masks of all 1s, but I'm not going to
 * superoptomize this stuff unless it's nessisary.
 */
inline static int maskcmp(const char *a1, const char *a2, int count,
	const char *mask)
{
	while (count) {
		if ((*a1 & *mask) != (*a2 & *mask))
			return 1;
		a1++, a2++, count--, mask++;
	}
	return 0;
}

/* Try to match one particular type of identifying information.
 */
inline static int try_match(char *data, int data_size,
	struct identify_info *i, int mode)
{
	int best_priority, match = 0, x;

	best_priority = 0;
	if (!data) return 0;
	if (!data_size) data_size = strlen(data);
	while (i) {
		switch (mode) {
			case COMPARE_STRING:
				match = (data_size == i->length)
					&& (!maskcmp(data, i->data,
						i->length, i->mask));
				break;
			case COMPARE_START:
				match = (data_size >= i->length)
					&& (!maskcmp(data, i->data,
						i->length, i->mask));
				break;
			case COMPARE_ALL:
				match = 0;
				for (x = 0; x < (data_size - i->length);
						x++) {
					if (!maskcmp(data + x, i->data,
						i->length, i->mask))
							match = 1;
						break;
				}
		}
		if (match && (i->priority > best_priority))
			best_priority = i->priority;
		i = i->next;
	}
	return best_priority;
}


/* This function will return the best match to a given file's type.  What's
 * particularily wierd is that the mime type and file extension are null
 * terminated, while the data has a seprate count.  Inconsistent, but
 * I think it works fine.
 */
struct type_info *E_scape_ident_find_match(char *mime, char *ext,
	char *data, int data_size)
{
	struct type_info *best, *i = info;
	int best_priority, mime_priority, ext_priority, magic_priority,
		string_priority, max_priority;

	best = 0;
	best_priority = -1;

	while (i) {
		mime_priority = try_match(mime, 0, i->mime,
			COMPARE_STRING);
		ext_priority = try_match(ext, 0, i->ext,
			COMPARE_STRING);
		magic_priority = try_match(data, data_size, i->magic,
			COMPARE_START);
		string_priority = try_match(data, data_size, i->string,
			COMPARE_ALL);
		max_priority = MAX(MAX(mime_priority, ext_priority),
			MAX(magic_priority, string_priority));
		if (max_priority > best_priority) {
			best = i;
			best_priority = max_priority;
		}
		i = i->next;
	}

	return best;
}
