/*   E-scape libimage switch code
     Copyright (C) 1997 Free Software Foundation, Inc.
     
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
     as published by the Free Software Foundation; either version 2
     of the License, or (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

/* This file trys to glue the format readers together with the input source
 * and output format.  There's a lack of flexibility in output formats at
 * the moment, but my guess is that if I ever write an image editor, it
 * will use a custom format for internal manipulation and translate to
 * XImage for output.  If this happens, I shouldn't have to change anything
 * except switch.c and libimage.h
 */

#include <config.h>
#include "image.h"

/* The following section will have to be changed when we stop using XImage
 */

/* called after dimensions have been determined
 */
void image_init(struct image_struct *image)
{
	char *data;
	int x, y;

	data = malloc(image->x * image->y * (image->depth / 8));
	image->data = XCreateImage(image->display, image->visual,
                image->depth, ZPixmap, 0, data, image->x, image->y, 8, 0);
	for (x = 0; x < image->x; x++)
		for (y = 0; y < image->y; y++)
			image_put_pixel_rgb(image, x, y, 0, 0, 0, 1);
}

void image_free(struct image_struct *image)
{
	XDestroyImage(image->data);
	free(image->private);
	free(image);
}

/* Prototype to keep things glued togther when this file is in the wrong
 * order
 */
static void unknown_addbytes(struct image_struct *image,
	char *data, size_t data_size);


/* I think this is the main function called by apps.
 */
struct image_struct *image_new(void *visual, void *display, int depth,
	int colormap)
{
	struct image_struct *tmp;

	tmp = malloc(sizeof(struct image_struct));
	tmp->data = 0;
	tmp->x = 0;
	tmp->y = 0;
	tmp->private = 0;
	tmp->addbytes = unknown_addbytes;
	tmp->visual = visual;
	tmp->display = display;
	tmp->depth = depth;
	tmp->colormap = colormap;
	return tmp;
}
/* This marks the conclusion of XImage specific code, I hope
 */

/* The following is a set of prototypes for identification functions for each
 * format.  If you want to add support for another format, simply add a
 * prototype immediately below this comment, and change unknown_addbytes
 * to call your function.
 */

/* image_identify_* will return 0 if the data is not a format it recognizes,
 * and 1 if the format is recognized.
 */

int image_identify_ppm(char *data, size_t data_size);
int image_identify_gif(char *data, size_t data_size);

void image_addbytes_ppm(struct image_struct *image,
	char *data, size_t data_size);
void image_addbytes_gif(struct image_struct *image,
	char *data, size_t data_size);

/* End of format function prototype
 */

/* Hopefully never used.  However if the ID code messes up we might just
 * use it.
 */
static void devnull_addbytes(struct image_struct *image,
	char *data, size_t data_size)
{
}

/* addbytes routine used until file format known.
 */
static void unknown_addbytes(struct image_struct *image,
	char *data, size_t data_size)
{
	if (image_identify_ppm(data, data_size))
		image->addbytes = image_addbytes_ppm;
	else if (image_identify_gif(data, data_size))
		image->addbytes = image_addbytes_gif;
	else
		image->addbytes = devnull_addbytes;
	(*(image->addbytes))(image, data, data_size);
}


/* Function called to deliver new data to an image.
 */
void image_addbytes(struct image_struct *image, char *data,
	size_t data_size)
{
	(*(image->addbytes))(image, data, data_size);
}
