/*   E-scape libtext
     Copyright (C) 1997 Free Software Foundation, Inc.
     
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
     as published by the Free Software Foundation; either version 2
     of the License, or (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

/*	The text code tries to handle plain ASCII, HTML, gopher menus,
	word processing files, etc.  If it's more or less text, I hope
	this code can support it.
  
	One known problem is that something like
	<H1>This is an <EM>example</H1> of overlap</EM>
	is not handled well, because this code assumes that tags are cleanly
	nested.  Although this limitation is expected with HTML, work processing
	programs generally allow this sort of thing.  Probably conversion to
	<H1>This is an <EM>example</EM></H1><EM> of overlap</EM>
	would be an acceptable interpretation.                               */

/*	You must include <config.h> and <sys/types.h> before including
	this file to make sure size_t is defined.  */

/*	Style sheets are used to describe the attributes of text--things
	like size, bold, etc.  Please look at the style sheet specs at
	http://www.w3.org/pub/WWW/TR/WD-css1.html
	Consider that file a part of the comments in this one ;-)
	FIXME: it's now a PR instead of a WD, so the URL may have changed... */

/*	The priority field is used to determine which rules apply if several
	selectors match.  In general, the most selective rule matches,
	so that you can have a rule for all <EM> tags, and another rules for
	<EM> tags that have <H1> as a parent, and the right thing will happen.
  
	PRIORITY_IMPORTANT and PRIORITY_AUTHOR are boolean attributes;
	PRIORITY_IMPORTANT is true when a rule is explicitly marked as
	important, and PRIORITY_AUTHOR is true when the rule is supplied
	with a particular HTML page, instead of supplied with the browser.
  
	The other three, PRIORITY_ID, PRIORITY_CLASS, and PRIORITY_TAG,
	are each multiplied by a count of the selectors in each class.
	For example, for a selector "UL OL LI.red", there are three
	tags, and one class, so the priority would be
	3 * PRIORITY_TAG + 1 * PRIORITY_CLASS.
  
	The priority usage is adapted from the style sheet specs; in the
	specs it says to put the integers together like this, but it doesn't
	mention that sorting by author and important in the same number
	works.
  
	The priority calcuation constants below assume 32 bit int.  Actually,
	27 bits should work flawlessly, and 26 will work if the numbers
	are unsigned.  In any event, the constants can be redefined
	if nessisary.           */

/* FIXME this is probably flawed */

#define PRIORITY_INLINE		0xFFFFFFFF

#define PRIORITY_IMPORTANT	0x02000000
#define PRIORITY_AUTHOR		0x01000000
#define PRIORITY_ID		0x00010000
#define PRIORITY_CLASS		0x00000100
#define PRIORITY_TAG		0x00000001

/*	The following struct definitions hold the rules from style sheets.
	See the style sheet specs for more information.

	A style sheet will be loaded into a list of css_rule's, whose semantic
	structure follows the semantic structure of a style sheet.  For each
	HTML tag, there will be a struct css_properties which lists its
	properties

	Initially, the per-HTML tag will be filled in with attributes from
	the parents; but then units will be evaluated into pixels, etc,
	before redering occurs.  (FIXME: that's probably too vague) */

/* Some things form a continum; others are distinct modes.  Unless doing
   arithmetic on the values makes sense, each possible value for some
   property will be a distinct mode.

   Since new modes will probably be defined with time, it is best if
   code reacts to an unrecognized mode by assuming it is `auto' or
   `normal'.  */
struct css_length {
  enum {
    /*	Used when no value has been specified */
    LENGTH_MODE_UNDEFINED = 0,

    /* The literal value `normal'. */
    LENGTH_MODE_NORMAL = 1,

    /* The literal value `auto'. */
    LENGTH_MODE_AUTO,

    /*	absolute named units `small', `medium', `large', etc.  Use the
	`name' field of the union `value' below to specify which one.  */
    LENGTH_MODE_ABSOLUTE,

    /* The literal value `smaller'.  */
    LENGTH_MODE_SMALLER,

    /* The literal value `larger'. */
    LENGTH_MODE_LARGER,

    /* Percentage values, which are relative to the parent generally.
       The value will NOT be conerted to a decimal, so if you specify
       100% in the style sheet, a value of 100 will be stored in
       `value.value'.  */
    LENGTH_MODE_PERCENT,

    /* Like percentage, but the factor is inherited.  Used in line_height.
       Also note that the number never needs to be scalled by 100.  */
    LENGTH_MODE_NUMBER,

    /* The unit is specified relative to the height of the `x' character,
       either in the font of the parent (only in the case of font_size)
       or the font of the element. */
    LENGTH_MODE_EX,

    /* The hieght of `m', relative to the height of the parent font (when
       specifying font_size) or the element's font.  This is defined so
       that `1ex' will give you the same size.  */
    LENGTH_MODE_EM,

    /* inches */
    LENGTH_MODE_IN,

    /* centimeters */
    LENGTH_MODE_CM,

    /* millimeters */
    LENGTH_MODE_MM,

    /* points (1/72 inch) */
    LENGTH_MODE_PT,

    /* picas (1 pica = 12 points) */
    LENGTH_MODE_PC,

    /* pixels.  never written in style sheets; only used internally.  */
    LENGTH_MODE_PX,

    /* Literal values that can be specified for `background-position',
       `vertical-align', etc.  */
    /* LENGTH_MODE_CENTER is used for `vertical-align: middle'.  */
    LENGTH_MODE_TOP,
    LENGTH_MODE_BOTTOM,
    LENGTH_MODE_CENTER,
    LENGTH_MODE_LEFT,
    LENGTH_MODE_RIGHT,

    /* Literal values used for `vertical-align'.  */
    /* FIXME some of these may well be stupid and redundant.  */
    LENGTH_MODE_BASELINE,
    LENGTH_MODE_SUB,
    LENGTH_MODE_SUPER,
    LENGTH_MODE_TEXT_TOP,
    LENGTH_MODE_TEXT_BOTTOM
  } units;
#define LENGTH_MODE_MAX LENGTH_MODE_TEXT_BOTTOM

  /* The number provided right before the unit; or for LENGTH_MODE_ABSOLUTE,
     one of the LENGTH_FOOBAR constants below. */
  float value;

#define LENGTH_XX_SMALL -3
#define LENGTH_X_SMALL -2
#define LENGTH_SMALL -1
#define LENGTH_MEDIUM 0
#define LENGTH_LARGE 1
#define LENGTH_X_LARGE 2
#define LENGTH_XX_LARGE 3
};

struct css_color {
  /* The red, green, and blue values will either all three be -1,
     indicating that no rgb value was given, or they will all three
     be in the range 0..0xFFFF, where 0 is black and 0xFFFF is full
     intensity. */
  int red, green, blue;

  /* The name of the color, if given, else 0. */
  char *name;

  /* The pixel value, where applicable.  */
  unsigned long pixel;
};

/*	The struct css_properties lists all the atributes whose values are
	specified.  As a general rule, a null pointer or an integer of 0
	is used if the value is not specified.
  
	In general, FOO_DEFINED is used so that we can tell the difference
	between whether something was unspecified (in which case the field will
	be 0), or was explicitly set to `none' or an equavalent value, in which
	case the variable will contain FOO_DEFINED. */

/*	This struct stores style information.  There is generally one per
	rule in a CSS1 stylesheet; there is also one for each tag.

	However, in most cases with strings, the display code has to
	figure out what to do (often nothing) when it sees 0.

	The display code also needs to deal with colors.  That's because
	there's a function (FIXME: XtDefaultForeground) which makes it
	easy to handle in the display code.  */
struct css_properties {
  /* A list of possible names for the font family.  Use the first
     one in the list that you can find a match for.  */
  /* The display code may not find any valid name in this list;
     so it supports a fallback.  */
  char **font_family;

  /* Whether to italicize the font.  */
  enum font_style_enum {
    FONT_STYLE_NORMAL = 1,
    FONT_STYLE_ITALIC,
    FONT_STYLE_OBLIQUE
  } font_style;

  /* Whether to use small caps.  */
  enum font_variant_enum {
    FONT_VARIANT_NORMAL = 1,
    FONT_VARIANT_SMALL_CAPS
  } font_variant;

  /* The absolute font wieght.  A value in 100, 200, 300, ..., 900.
     (Yes, strangely enough only 9 different values are possible--
     don't ask me why, just read the specs...) */
  /* FONT_WEIGHT_LARGER and FONT_WEIGHT_SMALLER are encoded
     here because it simplifies parsing. */
  /* FIXME make sure that such inheritance is properly supported. */
#define FONT_WEIGHT_NORMAL	400
#define FONT_WEIGHT_BOLD	700
#define FONT_WEIGHT_BOLDER	1
#define FONT_WEIGHT_LIGHTER	2
  int font_weight;

  /* font size */
  struct css_length font_size;

  /* The following two properties will probably become inadaquate when
     we internationalize E-scape properly... */
  /* used to store an XFontStruct * */
  void *font_opaque;
  /* used to store an XFontStruct * for the smaller font in small-caps mode */
  void *font_opaque_small_caps;

  /* Foreground color for text */
  struct css_color color;

  /* Background color */
  struct css_color background_color;

  /* URL of background image */
  char *background_url;

  /* The code in Net.c fills this in before trying to display the tag
     if background_url is non-zero. */
  struct image_struct *background_image;

#define BACKGROUND_REPEAT_DEFINED	0x01
#define BACKGROUND_REPEAT_X		0x02
#define BACKGROUND_REPEAT_Y		0x04
  int background_repeat;

  enum background_scroll_enum {
    BACKGROUND_SCROLL = 1,
    BACKGROUND_FIXED
  } background_scroll;

  struct css_length background_position_x, background_position_y;

  /* space between words in addition to default space */
  struct css_length word_spacing;

  /* extra space between letters */
  struct css_length letter_spacing;

#if 0
	/*	This is used to keep track of the color of underlining, because
		if you have HTML something like
		<UNDERLINE>foo<RED>bar</RED>baz</UNDERLINE>
		(assuming those are real HTML tags...), you want the underline
		color to be consistent throughout the whole length of the
		underline.  (At least that's what the style sheet specs say...)
		So this will get the underline color from the parent where
		appropriate.  FIXME do we really need this?  */
	struct css_color underline_color;
#endif

#define TEXT_DECORATION_DEFINED		0x01
#define TEXT_DECORATION_UNDERLINE	0x02
#define TEXT_DECORATION_OVERLINE	0x04
#define TEXT_DECORATION_LINE_THROUGH	0x08
#define TEXT_DECORATION_BLINK		0x10
  int text_decoration;

  struct css_length vertical_align;

#if 0
	enum vertical_align_enum {
		VERTICAL_ALIGN_BASELINE = 1,
		VERTICAL_ALIGN_SUB,
		VERTICAL_ALIGN_SUPER,
		VERTICAL_ALIGN_TOP,
		VERTICAL_ALIGN_TEXT_TOP,
		VERTICAL_ALIGN_MIDDLE,
		VERTICAL_ALIGN_BOTTOM,
		VERTICAL_ALIGN_TEXT_BOTTOM,
		VERTICAL_ALIGN_PERCENTAGE
	} vertical_align;
#endif

  /*	Case changing performed right before rendering */
  enum text_transform_enum {
    TRANSFORM_NONE = 1,
    TRANSFORM_UPPERCASE,
    TRANSFORM_LOWERCASE,
    TRANSFORM_CAPITALIZE
  } text_transform;

  /* indentation on first line */
  struct css_length text_indent;

  enum text_align_enum {
    JUSTIFY_JUSTIFY = 1,
    JUSTIFY_LEFT,
    JUSTIFY_RIGHT,
    JUSTIFY_CENTER
  } text_align;

  /* `line_height' contains the actual value in a style sheet, and the
     actual value used in rendering.  `line_height_inherit' of a parent
     is copied to `line_height' of a child during inheritance.  This
     special treatment is required because of the `number' option.  */
  struct css_length line_height, line_height_inherit;

  struct css_length margin_top;
  struct css_length margin_bottom;
  struct css_length margin_left;
  struct css_length margin_right;

  struct css_length padding_top;
  struct css_length padding_bottom;
  struct css_length padding_left;
  struct css_length padding_right;

  struct css_border {
    struct css_color color;
    struct css_length width;

    enum border_style_enum {
      BORDER_STYLE_NONE = 1,
      BORDER_STYLE_DOTTED,
      BORDER_STYLE_DASHED,
      BORDER_STYLE_SOLID,
      BORDER_STYLE_DOUBLE,
      BORDER_STYLE_GROOVE,
      BORDER_STYLE_RIDGE,
      BORDER_STYLE_INSET,
      BORDER_STYLE_OUTSET
    } style;
  } border_top, border_bottom, border_left, border_right;

  struct css_length width;
  struct css_length height;

  /* Note that his property is actually named `float', but if
     I use that name, I have a conflict with the C datatype... */
  enum float_mode_enum {
    FLOAT_NONE = 1,
    FLOAT_LEFT,
    FLOAT_RIGHT
  } float_mode;

#define CLEAR_DEFINED		0x01
#define CLEAR_LEFT		0x02
#define CLEAR_RIGHT		0x04
  int clear;

  enum display_enum {
    DISPLAY_NONE = 1,
    DISPLAY_BLOCK,
    DISPLAY_INLINE,
    DISPLAY_LIST_ITEM
  } display;

  /* The URL for an image.  This can be from a SRC or DATA attribute.
     It won't come from anything else.  Code which draws a list will
     check both `list_style_image_url' and then `image_url'.  */
  char *image_url;

  /* The actual image. */
  struct image_struct *image_image;

  /* URL for a list image.  This gets inherited into the image_url field.
     IMHO it would have been cleaner if the semantics were defined such
     that whatever is refered to by SRC comes before the actual text;
     and you would write `UL.LI { list-style: url(foo) }' instead of
     `UL { list-style: url(foo) }'; but we'red stuck with what we get... */
  char *list_style_image_url;

  /* The list marker image. */
  struct image_struct *list_style_image_image;

  enum list_style_type_enum {
    LIST_STYLE_NONE = 1,
    LIST_STYLE_DISC,
    LIST_STYLE_CIRCLE,
    LIST_STYLE_SQUARE,
    LIST_STYLE_DECIMAL,
    LIST_STYLE_LOWER_ROMAN,
    LIST_STYLE_UPPER_ROMAN,
    LIST_STYLE_LOWER_ALPHA,
    LIST_STYLE_UPPER_ALPHA
  } list_style_type;

  enum list_style_position_enum {
    LIST_STYLE_INSIDE = 1,
    LIST_STYLE_OUTSIDE
  } list_style_position;

  enum white_space_enum {
    WHITESPACE_NORMAL = 1,
    WHITESPACE_PRE,
    WHITESPACE_NOWRAP
  } white_space;

  /* Stored factors so that a child `font_size' will be able to read them
     if necissary. */
  float em_factor, ex_factor;
};

struct css_selector {
  /* The priority field contains the priority information for this
     struct css_selector as well as its parent, grandparent, etc.
     The PRIORITY_FOOBAR macros are used.  The rule associated with
     the struct css_selector which has the highest priority will be used.  */
  int priority;

  /* The tag field would point to "A" if you have "<A HREF=foobar>"
     in your HTML.  It would point to "BOLD" if you write "<BOLD>".  */
  char *tag;

  /* A vector of classes specified by the CLASS= attribute.  A null pointer
     marks the end of the vector; each string is null-terminated. */
  char **class;

  /*	The ID field contains an identifier that's unique to
	and appears in at most one place in a document. */
  char *id;

  /* A vector of pseudoclasses used for this selector.  */
  char **pseudo;

  /* Points to part of the selector for parent tags.  */
  struct css_selector *parent;

  /* Points to other selectors which are alternatives.  */
  struct css_selector *alt;
};


/*	A css_rule contains one rule in a style sheet; that is, it contains
	a list of all the cases where a collection of stylistic attributes
	apply.  For example, the following rule in a style sheet would be
	represented by one struct css_rule.

	H1, H2, H3 { color: red; font-family: helvetica }
 */

struct css_rule {
	/*	A list of all the selectors; either sel or sel->next or
		sel->next->next etc can apply, and make this css_rule
		be used.  (That assumes of course that you have a non-null
		sel->next...) */
	struct css_selector *sel;

	/*	A structure containing all the properities implied by this
		particular rule */
	struct css_properties *properties;

	/*	A link to the next css_rule in the chain.  */
	struct css_rule *next;
};

struct css_sheet {
  /* The style information in the sheet. */
  struct css_rule *rules;

  /* A pointer to the TITLE attribute for the style sheet, if one
     was provided in the LINK tag.  Note that this is NOT a copy;
     calling `free' for it will cause lossage.  Likewise, leaving
     this structure around after deleting the HTML tree will cause
     lossage. */
  char *title;

  /* If no title is given, `enable' will be 0.  If this is the first
     LINK with a given TITLE, then this is a pointer to its own
     enable_flag.  Otherwise, it will be a pointer to the enable_flag
     of the first item with the same title.  *enable will be 1
     if the style sheet should be used, and 0 if it shouldn't be used.  */
  int *enable, enable_flag;

  /* The next style sheet in the list. */
  struct css_sheet *next;
};

/*	Below are the data structures which describe the actual text.  
	An HTML document will always be expressed internally with this
	nesting style:
	<HTML><BODY><P>This is a short <strong>paragraph</strong>
	</P></BODY></HTML>.
	(Though internally the closing tags are not stored)
	What you should notice in that is that the toplevel tag
	is <HTML>, the second level tag is <BODY> (there's also
	such a thing as <HEAD>...), and then block level tags
	such as <P> and <H1> at the third level, with inline tags
	below that.  The CSS1 formatting model might be informative
	here.	*/

/*	An HTML tag is written something like
	<A HREF="http://www.gnu.ai.mit.edu/">
	The name field of the text_tag will be "A" for the sample above, and
	it has one item for the attr field: a tag_attribute with a name
	of "HREF" and a data field of "http://www.gnu.ai.mit.edu".  */

struct tag_attribute {
	char *name;
	char *data;
	struct tag_attribute *next;
};


/* A text_chunk can be either text or a tag.  A tree of text_chunk
   nodes describe a complete HTML document, along with some information
   on how to format it.  Each text_chunk either contains raw text
   or contains a tag.  */
struct text_chunk {
  /* The next `struct text_chunk' at the same level of nesting. */
  struct text_chunk *next;

  /* The previous `struct text_chunk' at the same level of nesting. */
  struct text_chunk *previous;

  /* The parent `struct text_chunk'. */
  struct text_chunk *parent;

  /* A child `struct text_chunk'.  This field will always be 0 if
     `data' is non-zero. */
  struct text_chunk *child;

  /* A pointer to the null-terminated name of the current tag,
     or zero if this `struct text_chunk' contains text. */
  char *tag_name;

  /* If `tag_name' is non-zero, a pointer to the attributes of the
     tag (if any); else zero. */
  struct tag_attribute *attr;

  /* A pointer to the first byte of text.  There will be `data_size'
     bytes of text.  If this field is zero, then `tag_name' must be
     non-zero. */
  char *data;

  /* The number of bytes pointed to by `data'. */
  size_t data_size;

  /* When `data' is zero, this will contain the properties to use where
     no pseudo-classes apply. */
  struct css_properties *properties;

  /* Zero if `properties' should be used, else a pointer to the properties
     to be used for the `first-line' pseudo-class. */
  struct css_properties *first_line_properties;

  /* Zero if `properties' should be used, else a pointer to the properties
     to be used for the `first-letter' pseudo-class. */
  struct css_properties *first_letter_properties;

  /* Zero if `properties' should be used, else a pointer to the properties
     to be used for the `marker' pseudo-class. */
  struct css_properties *marker_properties;

  /* Pointer to a widget where nessisary; else 0 */
  void *widget;
};

/* One line of text; or one block-level image. */
struct line_struct {
  /* The location of the top-left hand corner of the line. */
  int x1, y1;

  /* The location of the lower-right hand corner of the line. */
  int x2, y2;

  /* A pointer to the first `text_chunk' present in this line. */
  struct text_chunk *start_chunk;

  /* A pointer to the last `text_chunk present in this line. */
  struct text_chunk *end_chunk;

  /* A pointer to the first character present in this line, or 0
     if this `line_struct' represents something other than text. */
  char *start_char;

  /* A pointer to the character after the last character present in this
     line, or 0 if this `line_struct' represents something other than text. */
  char *end_char;

  /* The next line, according to the logical structure of the HTML markup. */
  struct line_struct *logical_next;
  /* The previous line, according to the logical structure of the HTML
     markup. */
  struct line_struct *logical_previous;

  /* The next physical line (sorted by y coordinate, and by x coordinate
     if the y coordinates are equal). */
  struct line_struct *physical_next;

  /* The previous physical line. */
  struct line_struct *physical_previous;
};


#if 0
/* FIXME change to a saner value */
#define DTD_HASH_SIZE 13

struct dtd_struct {
	char *name;
	enum dtd_type {
		CONTENT_EMPTY, /* No content */
		CONTENT_CDATA, /* Keep parsing until you find </foo> */
		CONTENT_NORMAL, /* allow arbitrary nested tags */
		CONTENT_OMIT, /* end tag might be omited; eg <LI> */
		CONTENT_PLAINTEXT /* keep parsing as text till EOF */
	} type;
	struct dtd_struct *next;
};
#endif

typedef struct html_parser_state {
	/*	pointer to the top-level html chunk, canonically refered to as
		<html>.  Note that this was based on the earlier
		theory that all documents would have <html>,
		<head>, and <body> tags inserted by the HTML parser.
		However, I decided that we WON'T insert an <html> chunk
		if there's none in the original file.  */
	struct text_chunk *html;

	/*	pointer to current chunk */
	struct text_chunk *current;

	/*	pointer to a future chunk; or a temporary pointer. */
	/*	FIXME shouldn't use as a temporary pointer for optomization */
	struct text_chunk *future;

	/* points to tag_attribute when parsing that section;
	   otherwise garbage */
	struct tag_attribute *tag_attr;

	/* used by fsmbuild */
	void *m;

	/*	1 if a new child chunk needs to be created under `current' */
	int new_child_chunk;

	/*	1 if a new next chunk needs to be created after `current' */
	int new_next_chunk;

	/*	used by Net.c to store widget pointer */
	void *self;

	void (*tag_start)(struct html_parser_state *x);

	void (*tag_finish)(struct html_parser_state *x);

	void (*line_finish)(struct html_parser_state *x);

} html_parser_state;

struct css1_parser_state {
	enum {
		STATE_TAG,
		STATE_CLASS,
		STATE_ID,
		STATE_PSEUDO
	} state;

	/* 1 if a new rule needs to be created */
	int new_rule;

	/* 1 if a new struct css_selector needs to be created as a parent */
	int new_parent;

	/* 1 if a new alterate struct css_selector needs to be created */
	int new_alt;

	struct css_rule *head_rule, *tail_rule;

	struct css_selector *sel;

	void *dest[4], *mapping;

	int integer[4];

	struct css_color color[4];

	struct css_length length[4];

	char *url;

	int part;

	float num, factor;

	char **vector;

	/* struct fsm_environment */
	void *m;

	/*	css_done and opaque used by whatever calls the parser  */
	void (*css_done)(struct css1_parser_state *x);

	void *opaque;
};

void parse_css1(struct css1_parser_state *x, char *data, size_t count);

/*	Initialize a new `struct css_properties' and initialized all its
	fields to appropriate values */
struct css_properties *new_css_properties (void);

/*	a version of new_css_properties that is customized for the toplevel
	element */
struct css_properties *default_css_properties (void);

/*	call free() for `foo' and all structures pointed to by `foo'  */
void free_css_rule(struct css_rule *foo);

/* Copy from `src' to `dest'.  If `use_factor' is non-zero, multiply
   the value by src->factor while copying it if appropraite.  */
void
copy_css_length(struct css_length *dest, struct css_length *src);

void copy_css_color(struct css_color *dest, struct css_color *src);

char **
copy_vector (char **src);

void clear_css_color(struct css_color *color);

void append_vector(char ***dest, char *src);

void free_vector(char **v);

struct html_parser_state *new_html_parser_state(void);

void parse_html(struct html_parser_state *x, char *data, size_t data_size);

/* return 1 if transformation is needed; return 0 if `do_transform'
   would just call `memcpy'. */
int
transform_needed (enum text_transform_enum text_transform);

/* copy `count' bytes of `source' to `destination', transforming them
   according to `text_transform'. */
void
do_transform (char *source, char *destintaion, size_t count,
	      enum text_transform_enum text_transform);

/* Given the `tag' and the parent elements it points to, calculate all
   style information needed to render the tag.  `content_style_type'
   must be a non-null pointer; by default it should be "text/css", although
   HTTP headers and meta tags can override this default, and this function
   will recognize overrides in `tag'.  `style_sheets' contains a list of
   style sheets to use to determine the style.

   `horizontal_factor' and `vertical_factor' specify the amount to multiply
   by to convert from millimeters to pixels for horizontal and vertical
   measurements, respectively.  `medium_base' is the number of pixels in
   a `medium'-sized font.  `medium_factor' is equal to `large' / `medium'.  

   After the font size and other related properties has been determined,
   `font_size_func' will be called, with `opaque' passed intact, to
   return the values `em_factor', `ex_factor', and `font_opaque'.

   `pseudo' is a vector giving the available pseudo-classes.
   `possible_pseudo_name' contains a vector of names for pseudo-classes
   which might be used in the current tag.  `possible_pseudo_used'
   is a vector of ints which will be set to 1 if a different value
   would be returned by this function if the corresponding element in
   `possible_pseudo_name' were provided in `psuedo'.  `possible_pseudo_count'
   contains the number of pseudo-classes specified by `possible_pseudo_name'.
   */
struct css_properties *
css_cascade_and_inherit (struct text_chunk *tag, char *content_style_type,
			 struct css_sheet *style_sheets,
			 void *opaque,
			 void (*font_size_func)(void *opaque,
						struct css_properties *properties,
						float *em_factor,
						float *ex_factor,
						void **font_opaque),
			 float horizontal_factor, float vertical_factor,
			 float medium_base, float medium_factor,
			 char **pseudo,
			 char **possible_pseudo_name,
			 int *possible_pseudo_used,
			 size_t possible_pseudo_count);
