/*   E-scape
     Copyright (C) 1997 Free Software Foundation, Inc.
     
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
     as published by the Free Software Foundation; either version 2
     of the License, or (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include <config.h>
#include "libit/alloca.h"
#include "noxt.h"
#include "libnet.h"
#include <stdio.h>
#include <signal.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/stat.h>

#if HAVE_FCNTL_H
#include <fcntl.h>
#endif

#if HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif

#include "libit/errno.h"
#include "libit/unistd.h"
#include "libit/string.h"
#include "libit/ctype.h"
#include "libit/dirent.h"

#include <libintl.h>
#define _(String) gettext(String)

#include "libit/malloc.h"

#ifndef FD_SETSIZE
#define FD_SETSIZE 256
#endif

static void input_data(struct net_connection *net, int *fd, XtInputId *id)
{
	char buf[65536];
	int count;

	count = read(*fd, &buf, 65536);
	if (count <= 0)
		_E_scape_net_done(net);
	else {
		bytes_transfered += count;
		net->current_size += count;
		while (net->current_size > net->total_size)
			net->total_size *= 2;
		net->filter->app_data(net->filter, (char *) &buf, count);
	}
}


void _E_scape_net_input_install(struct net_connection *net)
{
	net->data_id = XtAppAddInput(net->app_context, net->data_fd,
		net->read_condition, (XtInputCallbackProc) input_data, net);
}


void E_scape_net_stop(struct net_connection *net)
{
	/* Deal with the cache_fd later... */
	if (net->net_fd == net->data_fd)
		close(net->data_fd);
	else {
		if (net->net_fd != -1)
			close(net->net_fd);
		if (net->data_fd != -1)
			close(net->data_fd);
	}
	if (net->listen_fd != -1)
		close(net->data_fd);

	if (net->net_id)
		XtRemoveInput(net->net_id);
	if (net->data_id)
		XtRemoveInput(net->data_id);
	if (net->listen_id)
		XtRemoveInput(net->listen_id);
	if (net->write_id)
		XtRemoveInput(net->write_id);
	net->net_id = net->data_id = net->listen_id = net->write_id = 0;

	if (net->mime_encoding)
		free(net->mime_encoding);
	if (net->mime_type)
		free(net->mime_type);
	if (net->mime_field)
		free(net->mime_field);
	if (net->mime_info)
		free(net->mime_info);
	net->mime_encoding = net->mime_type = net->mime_field = net->mime_info
		= 0;

	if (net->ftp_buf)
		free (net->ftp_buf);
	net->ftp_buf = 0;

	if (net->location)
		free (net->location);
	net->location = 0;

	/* Don't explicitly close write_fd, since it simply mirrors
	 * another file descriptor.
	 */
	if (net->write_block_start)
		free(net->write_block_start);

	if (net->dns_pid)
		kill(net->dns_pid, SIGKILL);
	if (net->dns_out != -1)
		close(net->dns_out);
	if (net->dns_err != -1)
		close(net->dns_err);
	net->dns_err = net->dns_out = -1;
	if (net->dns_out_id)
		XtRemoveInput(net->dns_out_id);
	if (net->dns_err_id)
		XtRemoveInput(net->dns_err_id);
	net->dns_out_id = net->dns_err_id = 0;
}

void _E_scape_net_done(struct net_connection *net)
{
	net->filter->app_data(net->filter, 0, 0);
	E_scape_net_stop(net);
}

/* Make sure that E_scape_net_stop has been called on `net' before
   calling this function! */
void E_scape_net_free(struct net_connection *net)
{
	E_scape_net_stop(net);
	_E_scape_net_url_free (&net->url);
	if (net->auth_url)
		free(net->auth_url);
	if (net->auth_challenge)
		free(net->auth_challenge);
	if (net->auth_protocol)
		free(net->auth_protocol);
	if (net->auth_realm)
		free(net->auth_realm);
	if (net->location)
		free(net->location);
	free(net);
	/* A type of ISP...a freenet.  Now if only there was one in the .hi.us
	 * domain...sure there are nice sunsets, but they're getting dull.
	 *
	 * Since I have trouble being in the .hi.us domain, I just put myself
	 * in the .sub domain...
	 */
}

/* Currently, we only report the system error text if it is available.
 * I do not think that it is particularily essential, but it is nice
 * to have.  So I won't go out of my way to support it on older systems.
 *
 * Someday we may want to make a nice little table, complete with a style
 * sheet, and show how we parsed the URL, and what the IP address is, and
 * everything.  But for now, we won't bother.
 */
void _E_scape_net_html_error(struct net_connection *net, char *msg, int err)
{
	char *buf, *syserr;

/* As you can see below, religously commenting the #endif directive
 * serves no useful purpose.  Then again, neither does my complaining.
 */

#if HAVE_STRERROR
	if (err)
		syserr = strerror(err);
	else
#endif /* HAVE_STRERROR */
		syserr = "";

 	if ((!syserr) && err)
 	  {
	    /* 50 is an arbitrary number assumed to be large enough to
	       hold whatever integer size a system has. */
 	    syserr = alloca (50);
 	    sprintf (syserr, "System error %d", err);
 	  }


	net->mime_type = malloc(10);
	strcpy(net->mime_type, "text/html");
	buf = alloca(2 * strlen(msg) + strlen(syserr) + 25);
	sprintf(buf, "<TITLE>%s</TITLE><H1>%s</H1>%s",
		msg, msg, syserr);
	net->filter->app_data(net->filter, buf, strlen(buf));
	_E_scape_net_done(net);
}

static void lookup_error(struct net_connection *net, int *fd, XtInputId *id)
{
	int err, status;

	status = read(*fd, &err, sizeof(int));
	if (status < 0)
		_E_scape_net_html_error(net,
			_("Internal(?) Hostname Lookup Error"), errno);
	else
		_E_scape_net_html_error(net, _("Hostname Lookup Error"), err);
}

static void do_connect(struct net_connection *net, int *fd, XtInputId *id)
{
	int nonblock, status;
	struct sockaddr_in name;
	char *tmp_mesg;

	/* If socket messes up, then things are very bad.  Trying to generate
	 * a cute GUI error message probably would fail.
	 */
	net->data_fd = net->write_fd = net->net_fd =
		socket(AF_INET, SOCK_STREAM, 0);

	/* We try to make the socket non-blocking.  This may not be
	 * compatible with all systems--in libwww (used in lynx and
	 * mosaic), there is code for both fcntl and ioctl.  However,
	 * as far as I can tell fcntl is pretty standard.  So if someone
	 * out there needs ioctl or something else, please let me know.
	 *
	 * If you merely want it to compile, and don't care whether it
	 * works right, you may totally comment out the next 2 lines.
	 * Be aware that if a server is down E-scape will appear to freeze.
	 *
	 * I considered just saying #ifdef O_NONBLOCK, but I decided that
	 * that I want to hear from people if fcntl won't work.  So,
	 * if you are pretesting, and you "experience technical difficulties",
	 * please report this as a bug.
	 */
	nonblock = O_NONBLOCK;
	fcntl(net->net_fd, F_SETFL, &nonblock);
	/* If fcntl failed, then we'll just do the best we can. */

	read(*fd, &(name.sin_addr), sizeof(struct in_addr));
	name.sin_family = AF_INET;
	name.sin_port = htons(net->url.port);
	status = connect(net->net_fd, (struct sockaddr *) &name,
		sizeof(struct sockaddr_in));

	if ((status < 0) && (errno != EINPROGRESS) && (errno != EAGAIN))
		_E_scape_net_html_error(net, _("Connect Error"), errno);

	close(net->dns_out);
	close(net->dns_err);
	net->dns_err = net->dns_out = -1;
	net->dns_pid = 0;
	XtRemoveInput(net->dns_out_id);
	XtRemoveInput(net->dns_err_id);
	net->dns_out_id = net->dns_err_id = 0;

	if (net->status_func) {
	  /* FIXME: If the IP address were printed, that might be more
	     impressive */
	  tmp_mesg = alloca (256 + strlen (net->url.hostname));
	  sprintf (tmp_mesg, "Connecting to host %s, port %d\n",
		   net->url.hostname, net->url.port);
	  net->status_func (net, tmp_mesg);
	}

	net->net_id = XtAppAddInput(net->app_context, net->net_fd,
		net->write_condition,
		(XtInputCallbackProc) net->connection_done, net);
}

struct net_connection *E_scape_net_init(
	void (*app_data)(struct E_scape_filter *filter, char *buf,
		size_t count),
	void *app_context, void *read_condition, void *write_condition)
{
	struct net_connection *net;

	net = malloc(sizeof(struct net_connection));

	net->no_cache = 0;

	net->out_data = app_data;
	net->app_data = _E_scape_net_ident;
	net->status_func = 0;
	net->app_context = app_context;
	net->read_condition = read_condition;
	net->write_condition = write_condition;

	net->opaque = 0;

	net->mime_type = net->mime_encoding = 0;

	net->write_block_start = 0;
	net->write_fd = -1;
	net->write_id = 0;

	net->cache_fd = -1;
	net->net_fd = -1;
	net->data_fd = -1;
	net->listen_fd = -1;

	net->net_id = net->data_id = net->listen_id = net->write_id = 0;

	net->post_data = net->auth_post_data = 0;

	net->auth_challenge = net->auth_protocol = net->auth_realm = 
		net->location = net->auth_url = 0;

	_E_scape_net_url_zero (&net->url);

	net->dns_out = net->dns_err = -1;
	net->dns_pid = 0;
	net->dns_out_id = net->dns_err_id = 0;

	net->mime_field = net->mime_info = 0;
	net->mime_nl = ' ';
	net->mime_state = 0;

	net->ftp_buf = 0;

	net->filter = _E_scape_new_filter(net);

	return net;
}


/* FIXME localhost file handling really doesn't fit
 * in well with everything else.  Should probably
 * consider supporting MMAP...
 */
inline static void localhost_file(struct net_connection *net)
{
	struct stat stats;
	DIR *dir;
	struct dirent *ent;
	char perm[10];

	if (stat(net->url.filename, &stats)) {
		_E_scape_net_html_error(net, "stat", errno);
	} else if (S_ISREG(stats.st_mode)) {
		net->data_fd = open(net->url.filename, O_RDONLY);
		_E_scape_net_input_install(net);
	} else if (S_ISDIR(stats.st_mode)) {
		net->mime_type = strdup("text/x-name-dir");
		dir = opendir(net->url.filename);
		do {
			ent = readdir(dir);
			if (ent) {
#if 0
				filemodestring(&stats, &perm);
				net->app_data(net->filter, perm, 10);
				net->app_data(net->filter, " ", 1);
#endif
				net->app_data(net->filter, ent->d_name,
					strlen(ent->d_name));
				net->app_data(net->filter, "\n", 1);
			}
		} while (ent);
		closedir(dir);
		net->app_data(net->filter, 0, 0);
	} else {
		_E_scape_net_html_error(net, "Unknown file type", 0);
	}
}

void _E_scape_net_go(struct net_connection *net)
{
	int pipe_stdin[2], pipe_stdout[2], pipe_stderr[2], i;
	char *tmp_mesg;

	E_scape_net_stop(net);

	if (net->url.error) {
	  _E_scape_net_html_error (net, net->url.error, 0);
	  return;
	}

	/* FIXME url->fd */

	/* The following default is--BAD */
	net->total_size = 0xBAD;
	net->current_size = 0;

	if (!strncasecmp(net->url.protocol, "ftp", 3)) {
		if (net->url.port == -1)
			net->url.port = 21;
		net->connection_done = _E_scape_connected_ftp;
	} else if (!strncasecmp(net->url.protocol, "file", 4)) {
		if (strcasecmp(net->url.hostname, "localhost")) {
			if (net->url.port == -1)
				net->url.port = 21;
			net->connection_done = _E_scape_connected_ftp;
		} else {
			localhost_file(net);
			return;
		}
	} else if (!strncasecmp(net->url.protocol, "gopher", 6)) {
		if (net->url.port == -1)
			net->url.port = 70;
		net->connection_done = _E_scape_connected_gopher;
	} else if (!strncasecmp(net->url.protocol, "finger", 6)) {
		if (net->url.port == -1)
			net->url.port = 79;
		net->connection_done = _E_scape_connected_finger;
	} else if (!strncasecmp(net->url.protocol, "http", 4)) {
		if (net->url.port == -1)
			net->url.port = 80;
		net->connection_done = _E_scape_connected_http;
	} else if (!strncasecmp(net->url.protocol, "http_0_9", 8)) {
		if (net->url.port == -1)
			net->url.port = 80;
		net->connection_done = _E_scape_connected_http_0_9;
	} else {
		/* http is default */
		if (net->url.port == -1)
			net->url.port = 80;
		net->connection_done = _E_scape_connected_http;
	}

	pipe(pipe_stdin);
	pipe(pipe_stdout);
	pipe(pipe_stderr);
	net->dns_pid = fork();
	if (!net->dns_pid) {
		close(0);
		dup(pipe_stdin[0]);
		close(1);
		dup(pipe_stdout[1]);
		close(2);
		dup(pipe_stderr[1]);

		/* FIXME: It seems desireable to close all other
		 * file descriptors here.  This is not the best
		 * possible method, but it seems to work.
		 *
		 * I should say that it is not the most _CORRECT_
		 * possible method.  It is rather simple, and if
		 * it fails on a few rare occasions it probably won't
		 * matter...
		 */
		for (i = 3; i < FD_SETSIZE; i++)
			close(i);
		execl("/bin/sh", "sh", "-c", LIBEXECDIR "/e-scape-dns", 0);
		i = -1;
		write(2, &i, sizeof(int));
		exit(1);
	}
	close(pipe_stdin[0]);
	close(pipe_stdout[1]);
	close(pipe_stderr[1]);
	net->dns_out = pipe_stdout[0];
	net->dns_err = pipe_stderr[0];
	write(pipe_stdin[1], net->url.hostname, strlen(net->url.hostname) + 1);
	close(pipe_stdin[1]);
	net->dns_out_id = XtAppAddInput(net->app_context, net->dns_out,
		net->read_condition, (XtInputCallbackProc) do_connect, net);
	net->dns_err_id = XtAppAddInput(net->app_context, net->dns_err,
		net->read_condition, (XtInputCallbackProc) lookup_error, net);

	if (net->status_func) {
	  tmp_mesg = alloca (13 + strlen (net->url.hostname));
	  sprintf (tmp_mesg, "Looking up %s\n", net->url.hostname);
	  net->status_func (net, tmp_mesg);
	}
}

void E_scape_net_goto(struct net_connection *net, char *url, char *post_data,
	size_t post_size, int no_cache)
{
	/* FIXME: free memory.  Hopefully dmalloc will help */
	E_scape_net_stop(net);
	net->post_data = post_data;
	net->post_size = post_size;
	net->no_cache = no_cache;
	net->redir_count = 5;
	_E_scape_net_url_free(&net->url);
	_E_scape_net_url_parse(&net->url, url);
	_E_scape_net_go(net);
}
