/*   E-scape
     Copyright (C) 1997 Free Software Foundation, Inc.
     
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
     as published by the Free Software Foundation; either version 2
     of the License, or (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include <config.h>
#include "noxt.h"
#include "libnet.h"

#if HAVE_FCNTL_H
#include <fcntl.h>
#endif

#include <sys/types.h>
#include "libit/unistd.h"
#include "libit/string.h"
#include "libit/ctype.h"
#include "libit/malloc.h"

/* It may seem wasteful to not shrink the data, but for small amounts of data,
 * this is a trivial issue, and for sending files larger than physical RAM,
 * not resizing is essential for usable performace.  nautilus, which is my
 * primary development computer, has 8 MB of physical RAM, and currently
 * has 100 MB of swap.  (Notice the implication that the swap may go down
 * in the future, but I don't expect the physical RAM to go up...)
 * (Umm, actually, it turns out that nautilus now has 16 physical/
 *  100 virtual, so the above is completely wrong...)
 * If I use the following write code to send a 16 MB
 * file, the data will get swapped out once (I hope...), and then it will
 * be read back in as needed.
 *
 * Actually, if realloc() is brain damaged, it may get swaped through several
 * times, and writing 16 MB will take about 5 minutes.  In the case I describe
 * above performance will still be pathetic.  Probably if you want to upload
 * a local file, it should be read into RAM in 64K blocks and then written
 * out.  That would require a different interface to the generic write
 * function.  But that's a whole GNU project for another day.
 */

static void net_write(struct net_connection *net, int *fd, XtInputId *id)
{
	int count;

	count = write(net->write_fd, net->write_block_mark,
		net->write_block_end - net->write_block_mark);

#if 0
	write(2, net->write_block_mark, count);
#endif

	if (count == (net->write_block_end - net->write_block_mark)) {
		free(net->write_block_start);
		net->write_block_start = 0;
		XtRemoveInput(*id);
		net->write_id = 0;
	} else {
		net->write_block_mark += count;
	}
}

void _E_scape_net_write(struct net_connection *net, char *data, size_t count)
{
	char *old;

	if (net->write_block_start) {
		old = net->write_block_start;
		net->write_block_start = realloc(net->write_block_start, count
			+ (net->write_block_end - net->write_block_start));
		net->write_block_mark -= old - net->write_block_start;
		net->write_block_end -= old - net->write_block_start;
		memcpy(net->write_block_end, data, count);
		net->write_block_end += count;
	} else {
		net->write_block_start = net->write_block_mark = malloc(count);
		net->write_block_end = net->write_block_start + count;
		memcpy(net->write_block_start, data, count);
		net->write_id = XtAppAddInput(net->app_context, net->write_fd,
			net->write_condition, (XtInputCallbackProc) net_write,
			net);
	}
}
