/*   E-scape
     Copyright (C) 1997 Free Software Foundation, Inc.
     
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
     as published by the Free Software Foundation; either version 2
     of the License, or (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include <config.h>

#if HAVE_FCNTL_H
#include <fcntl.h>
#endif

#include "libit/unistd.h"
#include "libit/string.h"
#include "libit/ctype.h"
#include "noxt.h"
#include "libnet.h"

#include <libintl.h>
#define _(String) gettext(String)

#include "libit/malloc.h"

#define free_if_nonzero(foo) if (!foo) free (foo);

/* Return a pointer to a string which is a copy of the string starting
   at `start', and including all the characters from `start' to `stop'
   (inclusive).  The caller should evenually `free' the string. */
inline static char *
mydup (char *start, char *stop)
{
  int len = 1 + stop - start;
  char *tmp;

  tmp = malloc(len + 1);
  strncpy (tmp, start, len);
  tmp[len] = 0;
  return tmp;
}

/* Modify `dest', probably including reallocating it, according to
   the relative path pointed to by `start'.  `stop' points to the
   byte after the last byte of `start'. */
inline static void
relative_filename (char **dest, char *start, char *stop)
{
  char *last_slash;

  last_slash = *dest;
  while (*last_slash)
    last_slash++;
  while ((last_slash != *dest) && (*last_slash != '/'))
    last_slash--;
  while ((start[0] == '.')
	 && (start[1] == '.')
         && (start[2] == '/')
	 && (last_slash != *dest)
	 && ((start + 3) < stop)) {
    start += 3;
    last_slash--;
    while ((last_slash != *dest) && (*last_slash != '/'))
      last_slash--;
  }
  *dest = realloc (*dest, (last_slash - *dest) + (stop - start) + 1);
  memcpy (last_slash, start, stop - start);
  last_slash[stop - start] = 0;
}
      

/* Free data structures pointed to by `url' */
void
_E_scape_net_url_free (struct url *url)
{
  free_if_nonzero (url->url);
  free_if_nonzero (url->protocol);
  free_if_nonzero (url->username);
  free_if_nonzero (url->password);
  free_if_nonzero (url->hostname);
  free_if_nonzero (url->filename);
  free_if_nonzero (url->position);
  free_if_nonzero (url->error);
  if (url->fd != -1) {
    close (url->fd);
    url->fd = -1;
  }
  url->url = url->protocol = url->username = url->password
    = url->hostname = url->filename = url->position = url->error = 0;
}

void
_E_scape_net_url_zero (struct url *url)
{
  url->url = url->protocol = url->username = url->password
    = url->hostname = url->filename = url->position
    = url->error = 0;
  url->port = url->fd = -1;
}

/*
 * I'm hoping I can use > and < to compare string pointers.  Might not
 * work on wierd machines.  But pointer arithmetic is supposed to be
 * a standard part of C.
 *
 * Here's a map to try to clarify some of the varables and their locations:
 *
 * first_slash  passwd_start   at_sign   port_start
 *          |         |        |         |
 * protocol://username:password@host.name:port/nice/long/path#demo
 *         |  |                               |              |
 *     colon  host_area_start     host_area_end  path_area_end
 *
 * For the most part we point to the separator symbols because these
 * are easy to find.  At first host_area_start was the first actual
 * character in the username, until I found that caused problems comparing
 * it to at_sign.
 *
 * Note that this minimal URL should work if a host named "nautilus" is
 * running an httpd on port 80 in the local domain:
 *
 * nautilus
 */

/* Parse `URL', storing the result in `url'.  If `URL' should be treated
   as a URL relative to some other URL, then `url' should contain the
   url which it is relative to. */
void
_E_scape_net_url_parse (struct url *url, char *URL)
{
  char *colon, *first_slash, *host_area_start, *host_area_end,
    *path_area_start, *path_area_end, *at_sign, *passwd_start, *port_start;

  /* Nonzero if the URL may be in the form of `file://localhost/whatevers'.
     This is done because some web pages sometimes have URLs in the form
     of `file://localhost/dev/zero', and as far as I can see it generally
     doesn't make sense to allow such URLs unless they are entered by
     the user or come from another file://localhost URL. */
  int file_colon_slash_slash_localhost_allowed;

  file_colon_slash_slash_localhost_allowed = ((!url->protocol)
					      || (!url->hostname)
					      || (strcmp ("file",
							  url->protocol))
					      || (strcmp ("localhost",
							  url->hostname)));

  free_if_nonzero (url->url);
  url->url = strdup(URL);

  free_if_nonzero (url->error);
  url->error = 0;

  if ((!url->protocol) && (!url->hostname) && (!url->filename)) {
    path_area_end = strchr (URL, '#');
    if (path_area_end)
      url->filename = mydup (URL, path_area_end - 1);
    else
      url->filename = strdup (URL);
    url->fd = open (URL, O_RDONLY);
    if (url->fd != -1) {
      url->protocol = strdup ("file");
      url->hostname = strdup ("localhost");
      return;
    }
    free (url->filename);
    url->filename = 0;
  }

  colon = strchr (URL, ':');
  first_slash = strchr (URL, '/');
  if (colon
      && first_slash
      && (colon + 1 == first_slash)
      && (colon[2] == '/')) {
    if (url->protocol)
      free (url->protocol);
    url->protocol = mydup (URL, colon - 1);
    host_area_start = colon + 3;
  } else if (url->hostname) {
    if ((URL[0] == '/') && (URL[1] == '/') && (URL[2] != '/'))
      host_area_start = URL + 2;
    else
      host_area_start = 0;
  } else {
    host_area_start = URL;
    while (*host_area_start == '/')
      host_area_start++;
  }

  if (host_area_start) {
    url->port = -1;
    free_if_nonzero (url->username);
    free_if_nonzero (url->password);
    free_if_nonzero (url->hostname);
    free_if_nonzero (url->filename);
    free_if_nonzero (url->position);
    url->username = url->password = url->hostname = url->filename
      = url->position = 0;
    host_area_end = strchr (host_area_start, '/');
    if (!host_area_end) {
      host_area_end = host_area_start;
      while (*host_area_end)
	host_area_end++;
    }
    at_sign = strchr (host_area_start, '@');
    if (!at_sign || (host_area_end < at_sign))
      at_sign = host_area_start - 1;

    if ((host_area_start - 1) != at_sign) {
      passwd_start = strchr (host_area_start, ':');
      if (!passwd_start || (at_sign < passwd_start))
	passwd_start = at_sign;
      url->username = mydup (host_area_start, passwd_start - 1);
      if (passwd_start != at_sign) {
	url->password = mydup (passwd_start, at_sign);
      }
    }
    port_start = strchr (at_sign, ':');
    if (!port_start || (port_start > host_area_end))
      port_start = host_area_end;
    url->hostname = mydup (at_sign + 1, port_start - 1);
    if (port_start != host_area_end) {
      url->port = 0;
      port_start++;
      while (isdigit (*port_start)) {
	url->port *= 10;
	url->port += *port_start - '0';
	port_start++;
      }
    } else {
      url->port = -1;
    }
    path_area_start = host_area_end;
  } else {
    path_area_start = URL;
  }

  path_area_end = strchr (path_area_start, '#');
  if (!path_area_end) {
    path_area_end = host_area_end;
    while (*path_area_end)
      path_area_end++;
  }

  if (path_area_start == path_area_end) {
    if (!url->filename) {
      url->filename = realloc(url->filename, 2);
      url->filename[0] = '/';
      url->filename[1] = 0;
    }
  } else if (*path_area_start == '/') {
    free_if_nonzero (url->filename);
    url->filename = mydup (host_area_end, path_area_end - 1);
  } else {
    relative_filename (&url->filename, path_area_start, path_area_end);
  }

  free_if_nonzero (url->position);
  if (*path_area_end) {
    url->position = strdup (path_area_end + 1);
  } else {
    url->position = 0;
  }

  if ((!file_colon_slash_slash_localhost_allowed)
      && (!strcmp (url->protocol, "file"))
      && (!strcmp (url->hostname, "localhost"))) {
    url->error = strdup ("file://localhost URLs not allowed from here\n");
  }
}
