/*   E-scape
     Copyright (C) 1997 Free Software Foundation, Inc.
     
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
     as published by the Free Software Foundation; either version 2
     of the License, or (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include <config.h>
#include "pointermap.h"
#include "libit/malloc.h"

/* Insert a new item into `map'.  `x1' and `y1' specify the upper-left
   hand corner; `x2' and `y2' specify the lower-right hand corner.
   `callback' is the function to be called when a match is found;
   `opaque' is passed straight through to `callback', and `x' and `y'
   will contain the coordinates relative to `x1' and `y1'.
   `pointermap_insert' will be a no-op if an identical item is already
   present in `map'. */
void
pointermap_insert (struct pointermap **map, int x1, int y1, int x2, int y2,
		   void (*callback) (int x, int y, void *opaque),
		   void *opaque)
{
  struct pointermap *p;

  p = *map;

  while (p) {
    if ((p->x1 == x1) && (p->y1 == y1) && (p->x2 == x2) && (p->y2 == y2)
	&& (p->callback == callback) && (p->opaque == opaque))
      return;
    p = p->next;
  }
  p = malloc (sizeof (struct pointermap));
  p->x1 = x1;
  p->y1 = y1;
  p->x2 = x2;
  p->y2 = y2;
  p->callback = callback;
  p->opaque = opaque;
  p->next = *map;
  *map = p;
}
  

/* Call the callbacks in `map' for the regions in which which the point
   at `x' and `y' exists. */
void
pointermap_call (struct pointermap *map, int x, int y)
{
  while (map) {
    if ((x >= map->x1) && (x <= map->x2) && (y >= map->y1) && (y <= map->y2))
      map->callback (x - map->x1, y - map->y1, map->opaque);
    map = map->next;
  }
}

/* Adjust the coordinates in `map' by `x' and `y' pixels; delete objects
   which fall entirely in negative space, or entirely beyond `max_x'
   and `max_y'.  This is used when scrolling the screen. */
void
pointermap_move (struct pointermap **map, int x, int y, int max_x, int max_y)
{
  struct pointermap *previous, *current, *future;

  previous = 0;
  current = *map;

  while (current) {
    future = current->next;
    current->x1 += x;
    current->x2 += x;
    current->y1 += y;
    current->y2 += y;
    if (((current->x1 < 0) && (current->x2 < 0))
	|| ((current->y1 < 0) && (current->y2 < 0))
	|| ((current->x1 > max_x) && (current->x2 > max_x))
	|| ((current->y1 > max_y) && (current->y2 > max_y))) {
      if (previous)
	previous->next = future;
      else
	*map = future;
      free (current);
    }
    previous = current;
    current = future;
  }
}

/* Free `map' and `map->next' and `map->next->next' etc. */
void
pointermap_free (struct pointermap *map)
{
  struct pointermap *tmp;

  while (map) {
    tmp = map->next;
    free (map);
    map = tmp;
  }
}
