/*   E-scape zombie destrution code
     Copyright (C) 1997 Free Software Foundation, Inc.
     
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
     as published by the Free Software Foundation; either version 2
     of the License, or (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

/* I wish I had an example of a SIGCHLD handler that works reliably on
 * any system.  Of course that's going to be available about the same
 * time the whole world switches over to free software and Mr. Gates
 * becomes a homeless person.
 *
 * I think that on systems that don't have reliable signals, using unreliable
 * signals would be fine.  It may in strange circumstances leave an extra
 * zombie lying around once in a while, but I doubt it would be a problem.
 *
 * Also note that we are assuming that we want to ignore the exit status
 * of all programs.  Since the general philosophy in E-scape is that
 * we try to do our best to make things work, I think that is fine.
 *
 * Children fall into two categories: DNS processes and filters.  The
 * DNS program is written by me and I _KNOW_ I don't care about the
 * exit code.
 *
 * The user may have a filter which uses wierd return statuses when it
 * does its job correctly, so I think ignoring the status is rational.
 * Besides, we don't have to tell the user we messed up in, say, filtering
 * a graphic.  It will be painfully obvious from looking at the screen.
 *
 * On the other hand, if we can tell the user that filter XYZ messed up,
 * debugging might be easier.
 *
 * This also ignores SIGIO, since solaris likes to generate it and
 * kills my program when IO becomes available.
 */

#include <config.h>
#include <signal.h>
#include <sys/types.h>
#include "libit/unistd.h"
#include "libit/wait.h"

#ifndef SIGCHLD
#define SIGCHLD SIGCLD
#endif


static RETSIGTYPE handler()
{
	wait(0);
#if !HAVE_SIGACTION
	signal(SIGCHLD, handler);
#endif
}

static RETSIGTYPE sigio_handler()
{
#if !HAVE_SIGACTION
	signal(SIGIO, sigio_handler);
#endif
}

void E_scape_zombie_init(void)
{
#if HAVE_SIGACTION
	struct sigaction act;

	act.sa_handler = handler;
	sigemptyset(&act.sa_mask);
	act.sa_flags = 0;
	sigaction(SIGCHLD, &act, 0);

	act.sa_handler = sigio_handler;
	sigemptyset(&act.sa_mask);
	act.sa_flags = 0;
	sigaction(SIGIO, &act, 0);

#else /* HAVE_SIGACTION */
	signal(SIGCHLD, handler);
	signal(SIGIO, sigio_handler);
#endif
}
