/*   E-scape text case transformation routines
     Copyright (C) 1997 Free Software Foundation, Inc.
     
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
     as published by the Free Software Foundation; either version 2
     of the License, or (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include <config.h>
#include <ctype.h>
#include "libit/string.h"
#include "libtext.h"

/* copy `count' bytes of `source' to `destination', uppercasing them
   in the process. */
inline static void
uppercase (char *source, char *destination, size_t count)
{
  while (count) {
    *destination = toupper (*source);
    destination++, source++, count--;
  }
}

/* copy `count' bytes of `source' to `destination', lowercasing them
   in the process. */
inline static void
lowercase (char *source, char *destination, size_t count)
{
  while (count) {
    *destination = tolower (*source);
    destination++, source++, count--;
  }
}

/* copy `count' bytes of `source' to `destination', capitalizing them
   in the process. */
inline static void
capitalize (char *source, char *destination, size_t count)
{
  int alpha_seen = 0;

  while (count) {
    if (alpha_seen)
      *destination = *source;
    else
      *destination = tolower (*source);
    alpha_seen = isalpha (*source);
    destination++, source++, count--;
  }
}

/* copy `count' bytes of `source' to `destination', transforming them
   according to `text_transform'. */
void
do_transform (char *source, char *destination, size_t count,
	      enum text_transform_enum text_transform)
{
  switch (text_transform) {
  case TRANSFORM_UPPERCASE:
    uppercase (source, destination, count);
    break;
  case TRANSFORM_LOWERCASE:
    lowercase (source, destination, count);
    break;
  case TRANSFORM_CAPITALIZE:
    capitalize (source, destination, count);
    break;
  default:
    memcpy (destination, source, count);
    break;
  }
}

/* return 1 if transformation is needed; return 0 if `do_transform'
   would just call `memcpy'. */
int
transform_needed (enum text_transform_enum text_transform)
{
  switch (text_transform) {
  case TRANSFORM_UPPERCASE:
  case TRANSFORM_LOWERCASE:
  case TRANSFORM_CAPITALIZE:
    return 1;
  case TRANSFORM_NONE:
  }
  return 0;
}
