/*   E-scape
     Copyright (C) 1997 Free Software Foundation, Inc.
     
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
     as published by the Free Software Foundation; either version 2
     of the License, or (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

/* `draw_or_size_text' is used to either check the size of some text.

   `display' and `d' are used to specify where to draw; `gc' must be a
   valid gc, and its font and color are normally modified by this function.
   `gc_foreground' and `gc_font' are used to remember in the client
   what the current setting of the font and foreground is, so that server
   calls will only be generated when they need to change.

   `properties' is used to choose the style for this word.

   The text in pointed to by `data'; it is `data_size' long.  `data' should
   never have leading whitespace; but trailing whitespace should be
   passed to `draw_or_size_text'.  `draw_or_size_text' is normally
   called word by word, with no leading whitespace but with trailing
   whitespace, to find the size of words, so that the caller can figure
   out how many words will fit on one line.  The caller should then
   call this function with a whole line (if they whole line has
   the same `properties'; otherwise as large chunks of the line as possible)
   to actually draw the line; looking up the size should not involve
   round-trips to the X server, but actually drawing does involve
   round-trips to the X server.  However, note that with the
   :first-letter pseudo-class in CSS1, it is sometimes necissary
   to split a single word into multiple calls to `draw_or_size_text'.

   `actually_draw' should be non-zero if the word should actually
   be drawn to the screen; if it is non-zero, then this call should
   just find the dimensions and return.  However, the dimensions
   will be returned regardless of the state of `actually_draw'.

   `collapse_lws' should be non-zero if `draw_or_size_text' should
   collapse linear whitespace; it should be 0 if the linearwhitespace
   should be used unmangled.  `expand_tabs' should be 1 if a tab
   should be converted to 1 to 8 spaces, to make the tab line up
   properly; it should be zero if tabs should be left alone.
   One of `colapse_lws' or `expand_tabs' should be true; the
   semantics are not clean for having both of them or neither
   of them true.

   For each dimension, a 0 pointer is allowed, in which case the
   dimension will not be returned.  `ascent' and `descent' give the
   vertical dimensions of the word; `width' gives the width of the
   word itself.

   `x' and `y' contain the coordinates for the left size of the word
   at the baseline. */
void
draw_or_size_text (Display *display, Drawable d,
		   GC gc, unsigned long *gc_foreground, Font *gc_font,
		   struct css_properties *properties,
		   char *data, size_t data_size,
		   int x, int y,
		   int *ascent, int *descent, int *width,
		   int actually_draw, int collapse_lws, int expand_tabs);

/* The background drawing functions work on rectangular areas.
   Unfortunately, some paragraphs are non-rectangular.  So
   `draw_main_background' draws the central area where the image
   will be rendered, and `draw_aux_background' is used to draw other
   areas.  (If repeating of the image is enabled, however, the aux
   area can also have image in it.) */

/* Draw the background specified by `properties' in the box including
   (`x1', `y1') as its upper-left hand corner and (`x2', `y2')
   as its lower-right hand corner.  Draw to `display' and `d'
   using `gc'.  `gc_foreground' contains the current value of the
   foreground, and will be updated if `gc's foreground changes. */
void
draw_main_background (Display *display, Drawable d,
		      GC gc, unsigned long *gc_foreground,
		      struct css_properties *properties,
		      int x1, int y1, int x2, int y2);

/* Draw the background specified by `properties' in the box including
   (`x1', `y1') as its upper-left hand corner and (`x2', `y2')
   as its lower-right hand corner.  Draw to `display' and `d'
   using `gc'.  `gc_foreground' contains the current value of the
   foreground, and will be updated if `gc's foreground changes.
   (`x3', `y3') is the upper-left hand corner of the main region;
   (`x4', `y4') is the lower-right hand corner; they are used to
   align repeating images properly. */
void
draw_aux_background (Display *display, Drawable d,
		     GC gc, unsigned long *gc_foreground,
		     struct css_properties *properties,
		     int x1, int y1, int x2, int y2,
		     int x3, int y3, int x4, int y4);

