/*   E-scape
     Copyright (C) 1997 Free Software Foundation, Inc.
     
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
     as published by the Free Software Foundation; either version 2
     of the License, or (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include <config.h>
#include <sys/types.h>
#include <X11/Xlib.h>
#include "libit/ctype.h"
#include "libit/string.h"
#include "libtext.h"
#include "libit/malloc.h"

/* FIXME make PRIME a more useful value */
/*	PRIME should be 4093 or 1021.  Those are primes if I did my
	hand-checking on a toy computer correctly.  --nemo  */
#define PRIME 7

/*	Generate a hash value.  Case insensitive.  */
inline static int hash(char *s, int max)
{
	int n = 0;

	while (*s) {
		n += toupper(*s);
		s++;
	}

	return n % max;
}

inline static int color_hash(int r, int g, int b, int max)
{
	return (r * 7 + g * 5 + b * 3) % max;
}

struct name_cache_struct {
	char *name;
	int count;
	unsigned long pixel;
	struct name_cache_struct *next;
} *name_cache[PRIME];

struct rgb_cache_struct {
	int r, g, b;
	int count;
	unsigned long pixel;
	struct rgb_cache_struct *next;
} *rgb_cache[PRIME];



inline static int color_name_to_x_color(Display *display,
	Colormap colormap, char *name, unsigned long *pixel)
{
	int h, status;
	struct name_cache_struct *p;
	XColor result;

	h = hash(name, PRIME);

	p = name_cache[h];

	while (p) {
		if (!strcasecmp(p->name, name)) {
			p->count++;
			return p->pixel;
		}
		p = p->next;
	}

	status = XAllocNamedColor(display, colormap, name, &result, &result);

	if (!status)
		return 0;

	p = malloc(sizeof(struct name_cache_struct));

	p->count = 1;
	p->name = strdup(name);
	p->pixel = result.pixel;
	p->next = name_cache[h];
	name_cache[h] = p;

	*pixel = result.pixel;

	return 1;
}

inline static void free_name(Display *display, Colormap colormap, char *name)
{
	int h;
	struct name_cache_struct *p, *q;

	h = hash(name, PRIME);
	q = 0;
	p = name_cache[h];

	while (p) {
		if (!strcasecmp(name, p->name)) {
			XFreeColors(display, colormap, &(p->pixel), 1, 0);
			p->count--;
			if (!p->count) {
				if (q) {
					q->next = p->next;
					free(p);
				} else {
					name_cache[h] = p->next;
				}
			}
			return;
		}
		q = p;
		p = p->next;
	}
}

inline static int color_rgb_to_x_color(Display *display,
	Colormap colormap, int r, int g, int b, unsigned long *pixel)
{
	int h, status;
	struct rgb_cache_struct *p;
	XColor result;

	h = color_hash(r, g, b, PRIME);

	p = rgb_cache[h];

	while (p) {
		if ((p->r == r) && (p->g == g) && (p->b == b)) {
			p->count++;
			return p->pixel;
		}
		p = p->next;
	}

	result.red = r;
	result.green = g;
	result.blue = b;
	result.flags = DoRed | DoGreen | DoBlue;

	status = XAllocColor(display, colormap, &result);

	if (!status)
		return 0;

	p = malloc(sizeof(struct name_cache_struct));

	p->count = 1;
	p->r = r;
	p->g = g;
	p->b = b;
	p->pixel = result.pixel;
	p->next = rgb_cache[h];
	rgb_cache[h] = p;

	*pixel = result.pixel;

	return 1;
}

inline static void free_rgb(Display *display, Colormap colormap, int r, int g, int b)
{
	int h;
	struct rgb_cache_struct *p, *q;

	h = color_hash(r, g, b, PRIME);
	q = 0;
	p = rgb_cache[h];

	while (p) {
		if ((p->r == r) && (p->g == g) && (p->b == b)) {
			XFreeColors(display, colormap, &(p->pixel), 1, 0);
			p->count--;
			if (!p->count) {
				if (q) {
					q->next = p->next;
					free(p);
				} else {
					rgb_cache[h] = p->next;
				}
			}
			return;
		}
		q = p;
		p = p->next;
	}
}

int css1_color_to_x_color(Display *display, Colormap colormap,
	struct css_color *color, unsigned long *pixel)
{
	if (color->name) {
		return color_name_to_x_color(display, colormap, color->name,
			pixel);
	} else if ((color->red == -1)) {
		return color_rgb_to_x_color(display, colormap,
			color->red, color->green, color->blue, pixel);
	} else {
		return 0;
	}
}

void free_color(Display *display, Colormap colormap,
	struct css_color *color)
{
	if (color->name) {
		free_name(display, colormap, color->name);
	} else if ((color->red == -1)) {
		free_rgb(display, colormap, color->red, color->green,
			color->blue);
	}
}
