/*   E-scape
     Copyright (C) 1997 Free Software Foundation, Inc.
     
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
     as published by the Free Software Foundation; either version 2
     of the License, or (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

/*	This file takes a CSS1 font description and converts it to
	a vector of XFontStruct pointers.  Currently we assume that
	small-caps handling happens elsewhere.

	Two caches are kept.  One tracks XFontStruct matches by
	family, size, weight, and slant; the other keeps a list
	of all the fonts from the X server with a given name
	and slant.

	FIXME: No cleanup is ever done.  It would be worth looking into
	ways to make sure that there is a fixed amount of memory
	that can be allocated no matter what rogue style sheets
	people send to us.  The solution is probably not to cache
	non-matches.  */

#include <config.h>
#include "libit/alloca.h"
#include <sys/types.h>
#include <X11/Xlib.h>
#include <math.h>
#include <stdio.h>
#include "libit/string.h"
#include "libit/ctype.h"
#include "libtext.h"
#include "libit/malloc.h"

/* FIXME: add the five generic family names */
/* FIXME: use a default font when you don't find matches. */

/* FIXME make PRIME a more useful value */
/* PRIME should be 4093 or 1021.  Those are primes if I did my
   hand-checking on a toy computer correctly.  --nemo  */
#define PRIME 7

/* return a case insensitive hash value for `str', between 0
   and (max - 1) inclusive.  */
inline static int
hash(char *s, int max)
{
  int n = 0;

  while (*s) {
    n += toupper(*s);
    s++;
  }

  return n % max;
}

/* FIXME: The current wieght handling code only acknowleges the existance
   of medium and something darker than medium, since that seems
   to be all I can get with the fonts that happen to be installed
   on my machine.

   Anyone with better ideas is welcome to report it as a bug.  */

/* Extract the weight from the font description passed in `s', and
   return 0 for "medium", and 1 for anything else.  */
inline static int
extract_weight(char *s)
{
  s++;
  while (*s != '-') s++; s++;
  while (*s != '-') s++; s++;

  if ((s[0] == 'm') && (s[1] == 'e') && (s[2] == 'd') && (s[3] == 'i')
      && (s[4] == 'u') && (s[5] == 'm') && (s[6] == '-'))
    return 0;

  return 1;
}

/* Extract the font size from `s', and return as an integer */
inline static int
extract_size(char *s)
{
  int i;

  s++;
  while (*s != '-') s++; s++;
  while (*s != '-') s++; s++;
  while (*s != '-') s++; s++;
  while (*s != '-') s++; s++;
  while (*s != '-') s++; s++;
  while (*s != '-') s++; s++;

  i = 0;

  while ((*s >= '0') && (*s <= '9')) {
    i *= 10;
    i += *s - '0';
  }

  return i;
}

/* FIXME: Doug Rand says that it is sometimes the case that a fixed
   bitmap a few pixels off is better than a font scaled to the correct
   size.  Implement some support for this.  */
/* Look through the vector `q', and return a pointer to the string
   which most closely matches `size' and `weight'.  Write the chosen
   size to `out_size'.  If no match is found, return 0, leaving
   `out_size' undefined.  */
inline static char *
match(char *q, int size, int weight, int *out_size)
{
  /* The size which looks like the best choice so far, or -1 if no
     suitible match has been found. */
  int closest_size;

  /* The size and wieght of the font being examined. */
  int current_size, current_weight;

  /* The smallest and the largest font size that can be used.  */
  int min_size, max_size;

  /* The best medium and the best bold match. */
  char *medium, *bold;

  if (!q)
    return 0;

  medium = bold = 0;
  closest_size = -1;

  /* Use a 20% tolerance, which is the largest allowed by the CSS1
     spec.  FIXME maybe this should be user configurable? */
  max_size = size + (size / 5);
  min_size = size - (size / 5);

  while (q) {
    current_size = extract_size(q);
    current_weight = extract_weight(q);
    if (current_size == closest_size) {
      if (current_weight)
	bold = q;
      else
	medium = q;
    } else {
      /* Use an exact match wherever possible; otherwise use a scaleable
	 font if possible; otherwise use the closest match if it's
	 within min_size..max_size. */
      if ((current_size == size)
	  || ((current_size == 0) && (closest_size != size))
	  || ((closest_size != 0)
	      && ((abs(closest_size - size) > abs(current_size - size))
		  || (closest_size == -1))
	      && (current_size >= min_size) && (current_size <= max_size))) {
	bold = medium = q;
      }
    }
  }
  /* make sure that if a scalable font is used, the correct size is
     returned, so that the semantics of the code that depend on this
     work the way we want.  (the rest of the code could care less
     whether we scaled this font.) */
  if (closest_size)
    *out_size = closest_size;
  else
    *out_size = size;
  if (weight > 500)
    return bold;
  else
    return medium;
}

/* Given `display', `properties', `family' (the name of the font family),
   `font_size', and `style', return the best XFontStruct, and store
   the size actually choosen in `out_size'. */
static XFontStruct *
try_style(Display *display, struct css_properties *properties, char *family, int font_size,
	  enum font_style_enum style, int *out_size)
{
  char *s, w, *name;
  int h, count;
  struct cache_struct {
    char *s;
    char **list;
    struct cache_struct *next;
  } *p, *cache[PRIME];

  switch(style) {
  case FONT_STYLE_NORMAL:
    w = 'r';
    break;
  case FONT_STYLE_ITALIC:
    w = 'i';
    break;
  case FONT_STYLE_OBLIQUE:
    w = 'o';
    break;
  }

  /*	256 is a fudge factor.  */
  s = alloca(strlen(family) + 256);
  sprintf(s, "-*-%s-*-%c-*--*-*-*-*-*-*-*", family, w);

  h = hash(s, PRIME);

  p = cache[h];

  while (p) {
    if (!strcasecmp(p->s, s))
      break;
    p = p->next;
  }

  if (!p) {
    p = malloc(sizeof(struct cache_struct));
    p->s = strdup(s);
    /* FIXME there's an arbitrary limit below */
    p->list = XListFonts(display, s, 999999, &count);
    p->next = cache[h];
    cache[h] = p;
  }

  name = match(*(p->list), font_size, properties->font_weight, out_size);

  if (!name)
    return 0;

  return XLoadQueryFont(display, name);
}

/* Given `display', `properties', `family', and `font_size', find the best
   matching font, and return a pointer to its XFontStruct.  Return
   the choosen font size in `out_size'.  If no match is found, return
   0 and leave out_size undefined. */
static XFontStruct *
try_family(Display *display, struct css_properties *properties, char *family,
	   int font_size, int *out_size)
{
  XFontStruct *font;

  font = try_style(display, properties, family, font_size, properties->font_style,
		   out_size);

  if (font)
    return font;

  /* FIXME did I get the following right, or is it backwards? */
  if (properties->font_style == FONT_STYLE_ITALIC)
    return try_style(display, properties, family, font_size, FONT_STYLE_OBLIQUE,
		     out_size);

  return 0;
}

/* Try to find the appropriate XFontStruct in the cache; otherwise
   ask the X server what it is.
   `display' is a pointer to the X display for this font.
   `properties' is a pointer to the CSS propertiesibutes for the font being
   searched for.
   `family' is the name of the font family currently being tried.
   `font_size' is the input size.
   `actual_size' is an *output* parameter, which is the actual */
static XFontStruct *
try_cache(Display *display, struct css_properties *properties, char *family, int font_size,
	  int *actual_size)
{
  int h;

  struct cache_struct {
    /* name of the font family */
    char *family;

    /* size in pixels according to css1 input */
    int size;

    /* size of font according to X11 */
    int actual_size;

    enum font_style_enum style;

    int weight;

    XFontStruct *font;

    struct cache_struct *next;
  } *p, *cache[PRIME];

  h = (hash(family, PRIME) * 13 + font_size
       + properties->font_style * 1000 + properties->font_weight) % PRIME;

  p = cache[h];

  while (h) {
    if ((font_size == p->size)
	&& (properties->font_style == p->style)
	&& (properties->font_weight == p->weight)
	&& (!strcasecmp(family, p->family))) {
      *actual_size = p->actual_size;
      return p->font;
    }
    p = p->next;
  }

  p = malloc(sizeof(struct cache_struct));

  p->family = strdup(family);
  p->size = font_size;
  p->style = properties->font_style;
  p->weight = properties->font_weight;
  p->font = try_family(display, properties, family, font_size, &p->actual_size);
  p->next = cache[h];
  cache[h] = p;

  *actual_size = p->actual_size;
  return p->font;
}

/* Given the CSS1 font description in `properties', and `display', return a
   vector of possible XFontStructs.  The first one will be the best
   to use; the last the worst.  The reason the vector is needed at
   all is that some fonts lack some characters.

   `max_size' and `nominal_size' contain the largest size present and the
   size which should be used for 1em for the child font;
   `nominal_size' is the approximated size of the font if all
   the fonts in the vector are the same size; otherwise, it will
   be the input size.

   `font_size' should contain the desired size in pixels.  It is specified
   here so that two different fonts can be specified for one `properties' in
   order to make small-caps work. */
/* FIXME: make sure that at least one font is returned (or is this better
   to encode elsewhere?), and consider eliminating fonts if the consider
   no additional useful glyphs.  */
XFontStruct **
css1_font_to_x_font(Display *display, struct css_properties *properties, int font_size,
		    int *max_size, int *nominal_size)
{
  char **list;
  XFontStruct **font;
  int count;
  int current_size;

  font = malloc(sizeof(XFontStruct *));
  count = 0;
  list = properties->font_family;
  *nominal_size = *max_size = 0;

  while (*list) {
    font[count] = try_cache(display, properties, *list, font_size,
			    &current_size);
    if (font[count]) {
      count++;
      font = realloc(font, sizeof(XFontStruct *) * (count + 1));
      if (count == 1)
	*nominal_size = current_size;
      else if (*nominal_size != current_size)
	*nominal_size = font_size;
      if (current_size > *max_size)
	*max_size = current_size;
    }
    list++;
  }

  font[count] = 0;

  return font;
}






