/*   E-scape
     Copyright (C) 1997 Free Software Foundation, Inc.
     
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
     as published by the Free Software Foundation; either version 2
     of the License, or (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include <config.h>

#include "noxt.h"

#include "libit/unistd.h"

#include <sys/types.h>

/* For some reason, I am about to call my favorite kernel braindamaged.
 * WHY ARE THE FD_* MACROS in the _TIME_ header!?!?!?
 *
 * (I suspect if I asked on linux-kernel@vger.rutgers.edu, it would
 * be considered an extremely serious question, too technical for that
 * mailing list and either someone
 * would say something about the way other braindamaged systems do
 * things or they'd change it.  Oh well.)
 */

#include "libit/time.h"

#include "libit/errno.h"

#include <libintl.h>
#define _(String) gettext(String)

#include "libit/malloc.h"

/* The code in this directory provides some of the X Toolkit's functionality.
 * It is pretty bare bones, intended only to allow E-scape-get and E-scape-curses
 * to run.  You need to include <noxt.h> to use this.
 */

XtInputId XtAppAddInput(XtAppContext app, int source, XtPointer condition,
	XtInputCallbackProc proc, XtPointer call_data)
{
	_XtInputId *tmp;

#if 0
	fprintf(stderr, "Adding source %d\n", source);
#endif

	tmp = malloc(sizeof(_XtInputId));

	tmp->next = (_XtInputId *) app->inputs;
	if (tmp->next)
		tmp->next->back = tmp;
	tmp->back = 0;
	tmp->app = app;
	tmp->call_data = call_data;
	tmp->condition = condition;
	tmp->proc = proc;
	tmp->fd = source;

	app->inputs = (unsigned long) tmp;
	return app->inputs;
}

XtAppContext NoXtInit(void)
{
	XtAppContext tmp;

	tmp = malloc(sizeof(struct _XtAppContext));

	tmp->inputs = 0;

	return tmp;
}

void XtRemoveInput(XtInputId id)
{
	_XtInputId *tmp;

	tmp = (_XtInputId *) id;

#if 0
	fprintf(stderr, "Removing source %d\n", tmp->fd);
#endif

	if (tmp->back) {
		if (tmp->next)
			tmp->next->back = tmp->back;
		tmp->back->next = tmp->next;
		free(tmp);
	} else {
		if (tmp->next)
			tmp->next->back = 0;
		tmp->app->inputs = (unsigned long) tmp->next;
		free(tmp);
	}
}

#define CASTAND(foo, bar) (((int)(foo)) & ((int)(bar)))

void XtAppMainLoop(XtAppContext app)
{
	/* read, write, and exception file descriptor sets */
	fd_set rfd, wfd, efd;
	/* n is the largest fd + 1 */
	int n, status;
	_XtInputId *q;
	XtInputId id;

	while (1) {
		FD_ZERO(&rfd);
		FD_ZERO(&wfd);
		FD_ZERO(&efd);
		q = (_XtInputId *) app->inputs;
		n = 0;
		while (q) {
			/* I am _DEEPLY_ indepted to that wonderful
			 * idiot who thought that a number should
			 * be treated as a pointer.  Just to make
			 * sure that this code will have a certain
			 * degree of binary compabitility with the
			 * real thing, I continue this
			 * __braindamage__.
			 *
			 * POSIXLY_CORRECT was a serious GNU joke.
			 * Maybe I can convince them to add a
			 * __braindamage__ feature to gcc...nah, the
			 * free compiler isn't braindamaged.
			 */
			if (CASTAND(q->condition, XtInputReadMask)) {
				FD_SET(q->fd, &rfd);
			}
			if (CASTAND(q->condition, XtInputWriteMask)) {
				FD_SET(q->fd, &wfd);
			}
			if (CASTAND(q->condition, XtInputExceptMask)) {
				FD_SET(q->fd, &efd);
			}
			if (q->fd >= n)
				n = q->fd + 1;
			q = q->next;
		}
		status = select(n, &rfd, &wfd, &efd, 0);
		if (status < 0) {
			if (errno == EINTR)
				continue;
			else {
				perror(_("select (xt.c)"));
				exit(0);
			}
		}
		q = (_XtInputId *) app->inputs;
		while (q) {
			if ((CASTAND(q->condition, XtInputReadMask)
					&& FD_ISSET(q->fd, &rfd))
				|| (CASTAND(q->condition, XtInputWriteMask)
					&& FD_ISSET(q->fd, &wfd))
				|| (CASTAND(q->condition, XtInputExceptMask)
					&& FD_ISSET(q->fd, &efd))) {
				/* We must be careful, because the function
				 * we call may remove itself or other
				 * functions.  So we only call one function,
				 * and worry about the other file descriptors
				 * on the next loop.
				 */
					id = (unsigned long) q;
					q->proc(q->call_data, &(q->fd), &id);
					break;
				}
			q = q->next;
		}
	}
}
