/*   E-scape
     Copyright (C) 1997 Free Software Foundation, Inc.
     
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
     as published by the Free Software Foundation; either version 2
     of the License, or (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include <config.h>

#include "libit/alloca.h"

#include <math.h>
#include <X11/Xatom.h>
#include <X11/Xlib.h>

#include "libit/ctype.h"
#include "libtext.h"
#include "x.h"

/* return non-zero if `c' is linear whitespace; 0 if `c' is not lws */
inline static int
is_lws (char c)
{
  return (c == ' ') || (c == '\t') || (c == '\n') || (c == '\r');
}

/* return the width of `c' in `font' */
inline static int
extract_width (XFontStruct *font, char c)
{
  if (font->per_char)
    return font->per_char[c].width;
  else
    return font->max_bounds.width;
}

/* Return 1 if the `c' appears in `font'; else returns `0' */
inline static int
char_in_font (XFontStruct *font, char c)
{
  if ((c < font->min_char_or_byte2) || (c > font->max_char_or_byte2))
    return 0;
  c -= font->min_char_or_byte2;
  if (font->per_char[c].width)
    return 1;
  return 0;
}

/* FIXME we probably need to do some internationalization of the following
   function. */

/* Return the font to be used to render `c'.  `for_space' is prefered if
   c == ' '; otherwise `for_space' is ignored.  `font_for_char' then
   traverses the possibilities in `properties->font_opaque', and uses the
   first match.

   Returns the first font given if none match the character; hopefully
   this means that the default character in that font will be used
   and we won't get a crash. */
inline static XFontStruct *
font_for_char (struct css_properties *properties, char c,
	       XFontStruct *for_space)
{
  XFontStruct *mark;

  if ((c == ' ') && char_in_font (for_space, c))
    return for_space;

  if (islower (c) && properties->font_opaque_small_caps)
    mark = properties->font_opaque_small_caps;
  else
    mark = properties->font_opaque;
  while (mark) {
    if (char_in_font (mark, c))
      return mark;
    mark++;
  }
  return properties->font_opaque;
}

/* `draw_or_size_text' is used to either check the size of some text.

   `display' and `d' are used to specify where to draw; `gc' must be a
   valid gc, and its font and color are normally modified by this function.
   `gc_foreground' and `gc_font' are used to remember in the client
   what the current setting of the font and foreground is, so that server
   calls will only be generated when they need to change.

   `properties' is used to choose the style for this word.

   The text in pointed to by `data'; it is `data_size' long.  `data' should
   never have leading whitespace; but trailing whitespace should be
   passed to `draw_or_size_text'.  `draw_or_size_text' is normally
   called word by word, with no leading whitespace but with trailing
   whitespace, to find the size of words, so that the caller can figure
   out how many words will fit on one line.  The caller should then
   call this function with a whole line (if they whole line has
   the same style; otherwise as large chunks of the line as possible)
   to actually draw the line; looking up the size should not involve
   round-trips to the X server, but actually drawing does involve
   round-trips to the X server.  However, note that with the
   :first-letter pseudo-class in CSS1, it is sometimes necissary
   to split a single word into multiple calls to `draw_or_size_text'.

   `actually_draw' should be non-zero if the word should actually
   be drawn to the screen; if it is non-zero, then this call should
   just find the dimensions and return.  However, the dimensions
   will be returned regardless of the state of `actually_draw'.

   `collapse_lws' should be non-zero if `draw_or_size_text' should
   collapse linear whitespace; it should be 0 if the linearwhitespace
   should be used unmangled.  `expand_tabs' should be 1 if a tab
   should be converted to 1 to 8 spaces, to make the tab line up
   properly; it should be zero if tabs should be left alone.
   One of `colapse_lws' or `expand_tabs' should be true; the
   semantics are not clean for having both of them or neither
   of them true.

   For each dimension, a 0 pointer is allowed, in which case the
   dimension will not be returned.  `ascent' and `descent' give the
   vertical dimensions of the word; `width' gives the width of the
   word itself.

   `x' and `y' contain the coordinates for the left size of the word
   at the baseline. */
void
draw_or_size_text (Display *display, Drawable d,
		   GC gc, unsigned long *gc_foreground, Font *gc_font,
		   struct css_properties *properties,
		   char *data, size_t data_size,
		   int x, int y,
		   int *ascent, int *descent, int *width,
		   int actually_draw, int collapse_lws, int expand_tabs)
{
  /* The replacement_foo variables are pointed to by foo when foo is
     passed as 0, so that we don't have to worry about dealing with
     those variables being 0 later. */
  int replacement_ascent, replacement_descent, replacement_width;

  /* A temporary counter which contains the number of bytes in the current
     chunk (a `chunk' being the amount of text which is to be rendered
     in one XDrawString call; in other words, all of the consecutive text
     which should be rendered in the same font. */
  size_t chunk_size;

  /* The width of the current chunk in pixels. */
  int chunk_width;

  /* The font to use for the current chunk. */
  XFontStruct *chunk_font;

  /* The font to use for the next character. */
  XFontStruct *next_font;

  /* The current column number, starting from 0. */
  int current_column = 0;

  /* Column number of the first character of the current chunk, starting
     from 0. */
  int chunk_start_column;

  /* Set to 1 if the text needs to be mangled before calling XDrawString */
  int text_mangling_needed;

  /* temporary pointer to the mangled string */
  /* FIXME maybe this should be a more localized variable */
  char *mangled_string;

  /* The count of input bytes processed by the mangling code */
  size_t mangle_count;

  /* The destination pointer used during mangling */
  char *mangle_destination;

  /* True if the previous character was lws */
  int saw_lws;

  /* Number of pixels to add between characters */
  int letter_spacing;

  /* Number of pixels to add between words */
  int word_spacing;

  /* True when for some reason the current chunk must be terminated here */
  int chunk_done;

  if (!ascent)
    ascent = &replacement_ascent;
  if (!descent)
    descent = &replacement_descent;
  if (!width)
    width = &replacement_width;

  *ascent = *descent = *width = 0;

  if (properties->word_spacing.units == LENGTH_MODE_PX)
    word_spacing = rint (properties->word_spacing.value);
  else
    word_spacing = 0;

  if (properties->letter_spacing.units == LENGTH_MODE_PX)
    letter_spacing = rint (properties->letter_spacing.value);
  else
    letter_spacing = 0;

  if (actually_draw && (properties->color.pixel != *gc_foreground)) {
    XSetForeground (display, gc, properties->color.pixel);
    *gc_foreground = properties->color.pixel;
  }

  next_font = font_for_char (properties, *data,
			     ((XFontStruct *)properties->font_opaque));

  while (chunk_size != data_size) {
    chunk_font = next_font;
    chunk_start_column = current_column;
    text_mangling_needed = 0;
    saw_lws = 0;
    chunk_size = 1;
    chunk_width = extract_width (next_font, data[0]);
    chunk_done = 0;
    while ((chunk_size != data_size) && (next_font == chunk_font)) {
      if (expand_tabs && (data[chunk_size] == '\t')) {
	saw_lws = 1;
	next_font = font_for_char (properties, ' ', chunk_font);
	if (next_font == chunk_font) {
	  text_mangling_needed = 1;
	  chunk_size++;
	  chunk_width += extract_width (chunk_font, ' ') + letter_spacing;
	  current_column++;
	  while (current_column % 8) {
	    chunk_width += extract_width (chunk_font, ' ') + letter_spacing;
	    current_column++;
	  }
	}
	if (letter_spacing)
	  break;
      } else if (saw_lws && is_lws(data[chunk_size])) {
	chunk_size++;
	text_mangling_needed = 1;
      } else {
	next_font = font_for_char (properties, data[chunk_size], chunk_font);
	if (next_font == chunk_font) {
	  if ((properties->font_variant == FONT_VARIANT_SMALL_CAPS)
	      && (islower (data[chunk_size])))
	    text_mangling_needed = 1;
	  chunk_width += extract_width (chunk_font, data[chunk_size]);
	  chunk_size++;
	  current_column++;
	}
	saw_lws = is_lws (data[chunk_size]);
	if (letter_spacing)
	  chunk_width += letter_spacing;
	if (is_lws (data[chunk_size])) {
	  chunk_width += word_spacing;
	  break;
	}
	if (letter_spacing)
	  break;
      }
    }
    if (actually_draw) {
      if (chunk_font->fid != *gc_font) {
	XSetFont (display, gc, chunk_font->fid);
	*gc_font = chunk_font->fid;
      }
      if (text_mangling_needed) {
	/* The worst case is expanding each tab to 8 spaces. */
	mangled_string = alloca (chunk_size * 8);
	current_column = chunk_start_column;
	mangle_count = 0;
	mangle_destination = mangled_string;
	saw_lws = 0;
	while (mangle_count != chunk_size) {
	  if (expand_tabs && (data[mangle_count] == '\t')) {
	    saw_lws = 1;
	    current_column++;
	    *mangle_destination = ' ';
	    mangle_destination++;
	    while (current_column % 8) {
	      current_column++;
	      *mangle_destination = ' ';
	      mangle_destination++;
	    }
	  } else if (!saw_lws) {
	    if (properties->font_variant == FONT_VARIANT_SMALL_CAPS)
	      *mangle_destination = toupper (data[mangle_count]);
	    else
	      *mangle_destination = data[mangle_count];
	    mangle_destination++;
	    saw_lws = is_lws (data[mangle_count]);
	  }
	  mangle_count++;
	}
	XDrawString (display, d, gc, x, y, mangled_string,
		     mangle_destination - mangled_string); 
      } else {
	XDrawString (display, d, gc, x, y, data, chunk_size);
      }
    }
    x += chunk_width;
    width += chunk_width;
    data += chunk_size;
    data_size -= chunk_size;
    if (*ascent < chunk_font->ascent)
      *ascent = chunk_font->ascent;
    if (*descent < chunk_font->descent)
      *descent = chunk_font->descent;
  }
}
